"""
Command-line executable, run via ``python -m ...``.
"""
# Standard Library
import pathlib

from .cli import cli

if __name__ == '__main__':
    cli.main(prog_name=pathlib.Path(__file__).parent.name)



"""
Read and write the SAS XPORT/XPT file format from SAS Version 8 or 9.

The SAS V8 Transport File format, also called XPORT, or simply XPT, ...
"""

# Member names may now be up to 32 characters.
# Variable names may now be up to 32 characters.
# Variable labels may now be up to 256 characters.
# Character fields may now exceed 200 characters.
# Format fields may now exceed 8 characters (v9).

# Standard Library
import logging
import re
import struct
import warnings
from datetime import datetime

# Xport Modules
import xport.v56
from xport.v56 import _encoding, strftime, text_encode

__all__ = [
    'load',
    'loads',
    'dump',
    'dumps',
    '_encoding',
]

LOG = logging.getLogger(__name__)


class Library(xport.v56.Library):
    """
    Collection of datasets from a SAS Version 8 or 9 Transport file.
    """

    template = xport.v56.Library.template.replace(b'LIBRARY', b'LIBV8  ')
    pattern = re.compile(
        xport.v56.Library.pattern.pattern.replace(b'LIBRARY', b'LIBV8  '),
        re.DOTALL,
    )

    @classmethod
    def from_bytes(cls, bytestring):
        """
        Construct a ``Library`` from an XPORT-format bytes string.
        """
        return super().from_bytes(bytestring, MemberHeader=MemberHeader, Member=Member)

    def _bytes(self):
        return super()._bytes(Member=Member)


class MemberHeader(xport.v56.MemberHeader):
    """
    Dataset metadata from a SAS Version 8 or 9 Transport (XPORT) file.
    """

    pattern = re.compile(
        rb'HEADER RECORD\*{7}MEMBV8  HEADER RECORD\!{7}0{17}160{8}(?P<descriptor_size>140)  '
        rb'HEADER RECORD\*{7}DSCPTV8 HEADER RECORD\!{7}0{30}  '
        rb'SAS {5}(?P<name>.{32})SASDATA (?P<version>.{8})(?P<os>.{8})(?P<created>.{16})'
        rb'(?P<modified>.{16}) {16}(?P<label>.{40})(?P<type>    DATA|    VIEW| {8})'
        rb'HEADER RECORD\*{7}NAMSTV8 HEADER RECORD\!{7}0{6}(?P<n_variables>.{4})0{20}  '
        rb'(?P<namestrs>.*?)'
        rb'HEADER RECORD\*{7}LABELV(?P<v>8|9) HEADER RECORD\!{7}(?P<n_labels>.{5}) {27}'
        rb'(?P<labels>.*?)'
        rb'HEADER RECORD\*{7}OBSV8   HEADER RECORD\!{7}0{30}  ', re.DOTALL
    )

    @classmethod
    def from_bytes(cls, bytestring):
        """
        Construct a ``MemberHeader`` from an XPORT-format byte string.

        The Transport Version 8 format allows long variable names and labels.
        The Transport Version 9 format allows long variable format and informat
        descriptions.
        """
        self = super().from_bytes(bytestring, Namestr=Namestr)
        match = cls.pattern.search(bytestring)
        v9 = match['v'] == b'9'
        n = int(match['n_labels'].strip() or '0')
        data = match['labels']
        namestrs = {n.number: n for n in self.namestrs.values()}
        for _ in range(n):
            number, name_length, label_length = struct.unpack('>hhh', data[:6])
            i = (10 if v9 else 6) + name_length
            j = i + label_length
            namestrs[number].name = data[6:i].decode(xport.v56.TEXT_METADATA_ENCODING)
            namestrs[number].label = data[i:j].decode(xport.v56.TEXT_METADATA_ENCODING)
            if v9:
                format_length, informat_length = struct.unpack('>hh', data[6:10])
            data = data[j:]
            if v9:
                i = format_length
                j = i + informat_length
                namestrs[number].format = xport.Format.from_spec(data[:i].decode('ascii'))
                namestrs[number].informat = xport.Informat.from_spec(data[i:j].decode('ascii'))
                data = data[j:]
        if data and set(data) != {ord(b' ')}:
            raise ValueError(f'Expected only padding, got {data}')
        return self

    @classmethod
    def from_dataset(cls, dataset: xport.Dataset):
        """
        Construct a ``MemberHeader`` from an ``xport.Dataset``.
        """
        # The example "v8xpt" file I generated with Stata enumerated variables
        # starting from 0, which differs from the v5 xpt files I've seen.  This
        # may be a flaw in Stata's implementation that I've copied here.
        return super().from_dataset(dataset, Namestr=Namestr, variable_enumeration_start=0)

    template = f'''\
HEADER RECORD{'*' * 7}MEMBV8  HEADER RECORD{'!' * 7}{'0' * 17}16{'0' * 8}140  \
HEADER RECORD{'*' * 7}DSCPTV8 HEADER RECORD{'!' * 7}{'0' * 30}  \
SAS     %(name)32bSASDATA %(version)8b%(os)8b%(created)16b\
%(modified)16b{' ' * 16}%(label)40b%(type)8b\
HEADER RECORD{'*' * 7}NAMSTV8 HEADER RECORD{'!' * 7}{'0' * 6}\
%(n_variables)04d{'0' * 20}  \
%(namestrs)b\
HEADER RECORD{'*' * 7}LABELV8 HEADER RECORD{'!' * 7}%(n_labels)5s{' ' * 27}\
%(labels)b\
HEADER RECORD{'*' * 7}OBSV8   HEADER RECORD{'!' * 7}{'0' * 30}  \
'''.encode('ascii')

    def __bytes__(self):
        """
        Encode in XPORT format.
        """
        namestrs = b''.join(map(bytes, self.values()))
        if len(namestrs) % 80:
            namestrs += b' ' * (80 - len(namestrs) % 80)

        # TODO: Handle long format and informat names for Transport v9.
        labels = b''
        n_labels = 0
        triggers = {'label': 40}  # , 'format': 8, 'informat': 8}
        for namestr in self.values():
            strings = {
                'name': namestr.name,
                'label': namestr.label if namestr.label is not None else '',
            }
            strings = {k: v.encode(xport.v56.TEXT_METADATA_ENCODING) for k, v in strings.items()}
            if any(len(strings[k]) > l for k, l in triggers.items()):
                strings = list(strings.values())
                fmt = '>hhh' + ''.join(f'{len(s)}s' for s in strings)
                labels += struct.pack(fmt, namestr.number, *map(len, strings), *strings)
                n_labels += 1
        if len(labels) % 80:
            labels += b' ' * (80 - len(labels) % 80)

        return self.template % {
            b'name': text_encode(self, 'name', 32),
            b'label': text_encode(self, 'dataset_label', 40),
            b'type': text_encode(self, 'dataset_type', 8),
            b'n_variables': len(self),
            b'n_labels': str(n_labels or '').ljust(5).encode(xport.v56.TEXT_METADATA_ENCODING),
            b'os': text_encode(self, 'sas_os', 8),
            b'version': text_encode(self, 'sas_version', 8),
            b'created': strftime(self.created if self.created else datetime.now()),
            b'modified': strftime(self.modified if self.modified else datetime.now()),
            b'namestrs': namestrs,
            b'labels': labels,
        }


class Member(xport.v56.Member):
    """
    A dataset; a member of a dataset library.
    """

    @classmethod
    def from_bytes(cls, bytestring):
        """
        Parse a ``Member`` from an XPORT-format bytes string.
        """
        return super().from_bytes(bytestring, MemberHeader=MemberHeader)

    def _bytes(self):
        return super()._bytes(MemberHeader=MemberHeader)


class Namestr(xport.v56.Namestr):
    """
    Variable metadata from a SAS Version 8 or 9 Transport (XPORT) file.
    """

    # The C structure definition for namestr records in the v5 format
    # ends with 52 unused bytes.  The v8 format replaces that with:
    #
    #   char longname[32]   /* long name for Version 8-style */
    #   short lablen        /* length of label */
    #   char rest[18]       /* remaining fields are irrelevant */

    fmts = {
        140: '>hhhh8s40s8shhh2s8shhl32sh18s',
        # v5/6 had a 136-length option, but that is not supported in v8/9.
    }

    @classmethod
    def from_bytes(cls, bytestring: bytes):
        """
        Construct a ``Namestr`` from an XPORT-format byte string.
        """
        v56 = super().from_bytes(bytestring)
        fmt = cls.fmts[len(bytestring)]
        tokens = struct.unpack(fmt, bytestring)
        longname = tokens[-3].strip(b'\x00').decode(xport.v56.TEXT_METADATA_ENCODING
                                                    ).rstrip() or v56.name
        if v56.name not in longname:
            warnings.warn(f'Short name {v56.name} not included in {longname}')
        # TODO: What should I do with the label length?  ``tokens[-2]``
        return cls(
            vtype=v56.vtype,
            length=v56.length,
            number=v56.number,
            name=longname,
            label=v56.label,
            format=v56.format,
            informat=v56.informat,
            position=v56.position,
        )

    def __bytes__(self):
        """
        Encode in XPORT format.
        """
        LOG.debug(f'Encode {type(self).__name__}')
        fmt = self.fmts[140]

        longname = text_encode(self, 'name', 32)
        shortname = longname[:8]

        longlabel = (self.label if self.label is not None else '').encode(
            xport.v56.TEXT_METADATA_ENCODING
        )
        if len(longlabel) > 256:
            raise ValueError('ASCII-encoded label {longlabel} exceeds 256 characters')
        shortlabel = longlabel[:40].ljust(40)

        # TODO: What's the right way to handle long format names for v9?
        format_name = self.format.name.encode('ascii')[:8]
        informat_name = self.informat.name.encode('ascii')[:8]

        if self.number is None:
            raise ValueError('Variable number not assigned')
        if self.position is None:
            raise ValueError('Variable position not assigned')

        return struct.pack(
            fmt,
            self.vtype,
            0,  # "Hash" of name, always 0.
            self.length,
            self.number,
            shortname,
            shortlabel,
            format_name.ljust(8),
            self.format.length,
            self.format.decimals,
            self.format.justify,
            b'',  # Unused
            informat_name.ljust(8),
            self.informat.length,
            self.informat.decimals,
            self.position,
            longname,
            len(longlabel),
            b'',  # Padding
        )


def load(fp):
    """
    Deserialize a dataset library from a SAS Transport v8 (XPT) file.

        >>> with open('test/data/example.v8xpt', 'rb') as file:
        ...     library = load(file)
    """
    try:
        bytestring = fp.read()
    except UnicodeDecodeError:
        raise TypeError(f'Expected a BufferedReader in bytes-mode, got {type(fp).__name__}')
    return loads(bytestring)


def loads(bytestring):
    """
    Deserialize a dataset library from an XPORT-format string.

        >>> with open('test/data/example.v8xpt', 'rb') as file:
        ...     bytestring = file.read()
        >>> library = loads(bytestring)
    """
    return Library.from_bytes(bytestring)


def dump(library, fp):
    """
    Serialize a dataset library to a SAS Transport v8/9 (XPORT) file.

        >>> library = Library()
        >>> with open('test/data/doctest.v8xpt', 'wb') as file:
        ...     dump(library, file)

    The input ``library`` can be either an ``xport.Library``, an
    ``xport.Dataset`` collection, or a single ``pandas.DataFrame``.

        >>> ds = xport.Dataset(name='EMPTY')
        >>> with open('test/data/doctest.v8xpt', 'wb') as file:
        ...     dump(ds, file)
    """
    fp.write(dumps(library))


def dumps(library):
    """
    Serialize a dataset library to a string in XPORT format.

        >>> library = Library()
        >>> bytestring = dumps(library)
    """
    return bytes(Library(library))



"""
Read and write the CPORT file format.

CPORT-format is a compressed XPORT format.
"""

# Standard Library
from io import BytesIO


def load(fp):
    """
    Deserialize a SAS compressed transport file format document.

        with open('example.cpt', 'rb') as f:
            data = load(f)
    """
    raise NotImplementedError()


def loads(s):
    """
    Deserialize a SAS compressed transport file from a string.

        with open('example.cpt', 'rb') as f:
            bytestring = f.read()
        data = loads(bytestring)
    """
    fp = BytesIO(s)
    return load(fp)


def dump(columns, fp, name=None, labels=None, formats=None):
    """
    Serialize a SAS compressed transport file format document.

        data = {
            'a': [1, 2],
            'b': [3, 4],
        }
        with open('example.cpt', 'wb') as f:
            dump(data, f)
    """
    raise NotImplementedError()


def dumps(columns, name=None, labels=None, formats=None):
    """
    Serialize a SAS compressed transport file to a string.

        data = {
            'a': [1, 2],
            'b': [3, 4],
        }
        bytestring = dumps(data)
        with open('example.cpt', 'wb') as f:
            f.write(bytestring)
    """
    fp = BytesIO()
    dump(columns, fp)
    fp.seek(0)
    return fp.read()



"""
Tests for the core interface.
"""

# Standard Library
import math
import string
from io import BytesIO

# Community Packages
import pandas as pd
import pytest

# Xport Modules
import xport


class TestNaN:
    """
    Test special missing values.
    """

    def test_names(self):
        for c in '_' + string.ascii_uppercase:
            assert getattr(xport.NaN, c)

    def test_values(self):
        for c in '_' + string.ascii_uppercase:
            assert math.isnan(getattr(xport.NaN, c))


class TestInformat:
    """
    Verify parsing and display of input formats.
    """

    def test_spec_character(self):
        """
        Verify parsing from an informat specification.
        """
        assert xport.Informat.from_spec('$3.') == xport.Informat('$', 3, 0)
        assert xport.Informat.from_spec('$CHAR10.') == xport.Informat('$CHAR', 10, 0)

    def test_spec_numeric(self):
        """
        Verify parsing from an informat specification.
        """
        assert xport.Informat.from_spec('10.2') == xport.Informat('', 10, 2)
        assert xport.Informat.from_spec('dollar26.') == xport.Informat('DOLLAR', 26, 0)

    def test_struct_unpack(self):
        """
        Verify construction from struct tokens.
        """
        specs = [
            '$CHAR10.',
            'DOLLAR26.',
        ]
        for spec in specs:
            form = xport.Informat.from_spec(spec)
            b = bytes(form)
            cpy = xport.Informat.unpack(b)
            assert form.name == cpy.name
            assert form.length == cpy.length
            assert (form.decimals == cpy.decimals or form.decimals is None and cpy.decimals == 0)

    def test_display(self):
        """
        Verify string representation.
        """
        specs = [
            '$CHAR10.',
            'DOLLAR26.',
        ]
        for s in specs:
            iform = xport.Informat.from_spec(s)
            assert str(iform) == s


class TestFormat:
    """
    Verify parsing and display of variable formats.
    """

    def test_spec_character(self):
        """
        Verify parsing from an informat specification.
        """
        just = xport.FormatAlignment.LEFT
        assert xport.Format.from_spec('$3.', just) == xport.Format('$', 3, 0, just)
        just = xport.FormatAlignment.RIGHT
        assert xport.Format.from_spec('$CHAR10.', just) == xport.Format('$CHAR', 10, 0, just)

    def test_spec_numeric(self):
        """
        Verify parsing from an informat specification.
        """
        assert xport.Format.from_spec('10.2') == xport.Format('', 10, 2)
        assert xport.Format.from_spec('dollar26.') == xport.Format('DOLLAR', 26, 0)

    def test_display(self):
        """
        Verify string representation.
        """
        specs = [
            '$CHAR10.',
            'DOLLAR26.',
        ]
        for s in specs:
            iform = xport.Format.from_spec(s)
            assert str(iform) == s


class TestVariableMetadata:
    """
    Verify set/get and validation of variable metadata.
    """

    @staticmethod
    def compare_metadata(got, expected):
        for name in expected._metadata:
            assert getattr(got, name) == getattr(expected, name)
        assert got.vtype == expected.vtype

    def test_init(self):
        """
        Verify initialization.
        """
        v = xport.Variable(dtype='float')
        for name in v._metadata:
            getattr(v, name)  # Does not raise an error.

    def test_copy_metadata(self):
        """
        Verify ``Series`` methods that copy will keep SAS metadata.
        """
        v = xport.Variable(
            name='A',
            label='Alpha',
            format='$CHAR4.',
            dtype='string',
        )
        self.compare_metadata(v.copy(), v)
        self.compare_metadata(v.append(xport.Variable(['1'])), v)

    def test_format(self):
        v = xport.Variable(dtype='object')
        v.format = value = '$CHAR10.'
        assert v.format == xport.Format.from_spec(value)
        with pytest.raises(ValueError):
            v.format = ''
        with pytest.raises(ValueError):
            v.format = '$abcdefghi1.'

    def test_informat(self):
        v = xport.Variable(dtype='object')
        v.informat = value = '10.2'
        assert v.informat == xport.Informat.from_spec(value)
        with pytest.raises(ValueError):
            v.informat = '1.2.3'

    @pytest.mark.skip('Not implemented')
    def test_vtype(self):
        character = ['string', 'object']
        numeric = ['float', 'int', 'bool']
        invalid = ['datetime64[ns]']
        for dtype in character:
            v = xport.Variable(dtype=dtype)
            assert v.vtype == xport.VariableType.CHARACTER
        for dtype in numeric:
            v = xport.Variable(dtype=dtype)
            assert v.vtype == xport.VariableType.NUMERIC
        for dtype in invalid:
            with pytest.raises(TypeError):
                xport.Variable(dtype=dtype)


class TestDatasetMetadata:
    """
    Verify set/get and validation of dataset metadata.
    """

    @staticmethod
    def compare_metadata(got, expected):
        for name in expected._metadata:
            assert getattr(got, name) == getattr(expected, name)
        for k, v in expected.items():
            TestVariableMetadata.compare_metadata(got[k], v)
        assert (got.contents == expected.contents).all(axis=None)

    def test_init(self):
        """
        Verify initialization.
        """
        v = xport.Dataset()
        for name in v._metadata:
            getattr(v, name)  # Does not raise an error.

    def test_copy_metadata(self):
        """
        Verify ``DataFrame`` methods that copy will keep SAS metadata.
        """
        ds = xport.Dataset(
            data={
                'a': [1],
                'b': xport.Variable(['x'], label='Beta')
            },
            name='EXAMPLE',
            label='Example',
        )
        self.compare_metadata(ds.copy(), ds)
        self.compare_metadata(
            ds.append(pd.DataFrame({
                'a': [2],
                'b': ['y'],
            })),
            ds,
        )
        self.compare_metadata(pd.concat([ds, ds]), ds)

    def test_contents(self):
        """
        Verify variables metadata summary.
        """
        ds = xport.Dataset(
            data={
                'a': [1],
                'b': xport.Variable(['x'], label='Beta'),
                'c': [None],
            },
            name='EXAMPLE',
            label='Example',
        )
        ds['a'].vtype = xport.VariableType.NUMERIC
        ds['b'].vtype = xport.VariableType.CHARACTER
        got = ds.contents
        assert list(got.index) == [1, 2, 3]
        assert list(got['Label']) == ['', 'Beta', '']
        assert list(got['Type']) == ['Numeric', 'Character', '']


class TestLibrary:
    """
    Verify get/set of dataset library attributes and members.
    """

    def test_create_empty(self):
        xport.Library()

    def test_create_from_mapping(self):
        with pytest.warns(UserWarning, match=r'Set dataset name'):
            lib = xport.Library({'x': xport.Dataset()})
        assert 'x' in lib
        with pytest.raises(ValueError):
            xport.Library({'x': xport.Dataset(name='y')})

    def test_create_from_list(self):
        lib = xport.Library(xport.Dataset())
        assert None in lib
        with pytest.warns(UserWarning, match=r'More than one dataset named'):
            xport.Library([xport.Dataset(), xport.Dataset()])

    def test_create_from_dataframe(self):
        lib = xport.Library(pd.DataFrame())
        assert None in lib


class TestLegacy:
    """
    Verify deprecated API still works.
    """

    # Gotta stay backwards compatible.  The FDA has written docs.

    def test_from_columns(self, library):
        ds = next(iter(library.values()))
        mapping = {k: v for k, v in ds.items()}
        fp = BytesIO()
        with pytest.warns(DeprecationWarning):
            xport.from_columns(mapping, fp)
        fp.seek(0)
        result = next(iter(xport.v56.load(fp).values()))
        assert (result == ds).all(axis=None)

    def test_from_rows(self, library):
        ds = next(iter(library.values()))
        rows = list(ds.itertuples(index=None, name=None))
        fp = BytesIO()
        with pytest.warns(DeprecationWarning):
            xport.from_rows(rows, fp)
        fp.seek(0)
        result = next(iter(xport.v56.load(fp).values()))
        assert (result.values == ds.values).all(axis=None)

    def test_from_dataframe(self, library):
        ds = next(iter(library.values()))
        fp = BytesIO()
        with pytest.warns(DeprecationWarning):
            xport.from_dataframe(ds, fp)
        fp.seek(0)
        result = next(iter(xport.v56.load(fp).values()))
        assert (result == ds).all(axis=None)

    def test_to_rows(self, library, library_bytestring):
        ds = next(iter(library.values()))
        fp = BytesIO(library_bytestring)
        with pytest.warns(DeprecationWarning):
            result = xport.to_rows(fp)
        df = pd.DataFrame(result)
        assert (df.values == ds.values).all(axis=None)

    def test_to_columns(self, library, library_bytestring):
        ds = next(iter(library.values()))
        fp = BytesIO(library_bytestring)
        with pytest.warns(DeprecationWarning):
            result = xport.to_columns(fp)
        df = pd.DataFrame(result)
        assert (df == ds).all(axis=None)

    def test_to_numpy(self, library, library_bytestring):
        ds = next(iter(library.values()))
        fp = BytesIO(library_bytestring)
        with pytest.warns(DeprecationWarning):
            result = xport.to_numpy(fp)
        assert (result == ds.values).all(axis=None)

    def test_to_dataframe(self, library, library_bytestring):
        ds = next(iter(library.values()))
        fp = BytesIO(library_bytestring)
        with pytest.warns(DeprecationWarning):
            result = xport.to_dataframe(fp)
        assert (result == ds).all(axis=None)



"""
Tests for XPT format from SAS versions 5 and 6.
"""

# Standard Library
import math
import string
from datetime import datetime

# Community Packages
import pandas as pd
import pytest

# Xport Modules
import xport
import xport.v56


@pytest.fixture(scope='module')
def library():
    """
    Create a 4-column, 6-row dataset with numbers and text.
    """
    ds = xport.Dataset(
        data={
            'VIT_STAT': ['ALIVE'] * 3 + ['DEAD'] * 3,
            'ECON': ['POOR', 'NOT', 'UNK'] * 2,
            'COUNT': [1216, 1761, 2517, 254, 60, 137],
            'TEMP': [98.6, 95.4, 86.7, 93.4, 103.5, 56.7],
        },
        name='ECON',
        label='Blank-padded dataset label',
        dataset_type='',
    )
    ds.created = ds.modified = datetime(2015, 11, 13, 10, 35, 8)
    ds.sas_os = 'W32_7PRO'
    ds.sas_version = '9.3'
    ds['VIT_STAT'].label = 'Vital status'
    ds['VIT_STAT'].format = '$5.'
    ds['VIT_STAT'].informat = xport.Informat()
    ds['VIT_STAT'].width = 8
    ds['ECON'].label = 'Economic status'
    ds['ECON'].format = xport.Format('$CHAR', 4, 0, xport.FormatAlignment.RIGHT)
    ds['ECON'].informat = xport.Informat()
    ds['ECON'].width = 8
    ds['COUNT'].label = 'Count'
    ds['COUNT'].format = 'comma8.0'
    ds['COUNT'].informat = xport.Informat()
    ds['COUNT'].width = 8
    ds['TEMP'].label = 'Temperature'
    ds['TEMP'].format = '8.1'
    ds['TEMP'].informat = xport.Informat()
    ds['TEMP'].width = 8
    return xport.Library(
        members=[ds],
        created=ds.created,
        modified=ds.modified,
        sas_os=ds.sas_os,
        sas_version=ds.sas_version,
    )


@pytest.fixture(scope='module')
def library_bytestring():
    """
    Create the same dataset in SAS V5 Transport format.
    """
    return b'''\
HEADER RECORD*******LIBRARY HEADER RECORD!!!!!!!000000000000000000000000000000  \
SAS     SAS     SASLIB  9.3     W32_7PRO                        13NOV15:10:35:08\
13NOV15:10:35:08                                                                \
HEADER RECORD*******MEMBER  HEADER RECORD!!!!!!!000000000000000001600000000140  \
HEADER RECORD*******DSCRPTR HEADER RECORD!!!!!!!000000000000000000000000000000  \
SAS     ECON    SASDATA 9.3     W32_7PRO                        13NOV15:10:35:08\
13NOV15:10:35:08                Blank-padded dataset label                      \
HEADER RECORD*******NAMESTR HEADER RECORD!!!!!!!000000000400000000000000000000  \
\x00\x02\x00\x00\x00\x08\x00\x01VIT_STATVital status                            \
$       \x00\x05\x00\x00\x00\x00\x00\x00        \x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x02\x00\x00\x00\x08\x00\x02ECON    Economic status                         \
$CHAR   \x00\x04\x00\x00\x00\x01\x00\x00        \x00\x00\x00\x00\x00\x00\x00\x08\
\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x01\x00\x00\x00\x08\x00\x03COUNT   Count                                   \
COMMA   \x00\x08\x00\x00\x00\x00\x00\x00        \x00\x00\x00\x00\x00\x00\x00\x10\
\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x01\x00\x00\x00\x08\x00\x04TEMP    Temperature                             \
        \x00\x08\x00\x01\x00\x00\x00\x00        \x00\x00\x00\x00\x00\x00\x00\x18\
\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\
HEADER RECORD*******OBS     HEADER RECORD!!!!!!!000000000000000000000000000000  \
ALIVE   POOR    CL\x00\x00\x00\x00\x00\x00Bb\x99\x99\x99\x99\x99\x98\
ALIVE   NOT     Cn\x10\x00\x00\x00\x00\x00B_fffffh\
ALIVE   UNK     C\x9dP\x00\x00\x00\x00\x00BV\xb333334\
DEAD    POOR    B\xfe\x00\x00\x00\x00\x00\x00B]fffffh\
DEAD    NOT     B<\x00\x00\x00\x00\x00\x00Bg\x80\x00\x00\x00\x00\x00\
DEAD    UNK     B\x89\x00\x00\x00\x00\x00\x00B8\xb333334\
                                                \
'''


@pytest.fixture(scope='module')
def library_header_bytestring(library_bytestring):
    """
    First 3 lines of the file.
    """
    return library_bytestring[:80 * 3]


@pytest.fixture(scope='module')
def dataset_bytestring(library_bytestring, library_header_bytestring):
    """
    Everything after the library header.
    """
    return library_bytestring[len(library_header_bytestring):]


@pytest.fixture(scope='module')
def member_header_bytestring(dataset_bytestring):
    """
    First 5 lines of the member dataset, plus namestrs and a final line.
    """
    index = 80 * 5 + 140 * 4 + 80 * 1
    return dataset_bytestring[:index]


@pytest.fixture(scope='module')
def observations_bytestring(dataset_bytestring, member_header_bytestring):
    """
    Observations are streamed in the file after the member header.
    """
    return dataset_bytestring[len(member_header_bytestring):]


@pytest.fixture(scope='module')
def namestr_bytestring(member_header_bytestring):
    """
    Namestrs start after 5 header lines and are 140 bytes long.
    """
    index = 80 * 5
    return member_header_bytestring[index:index + 140]


@pytest.fixture(scope='module')
def variable(library):
    """
    Example variable.
    """
    return library['ECON']['VIT_STAT']


@pytest.fixture(scope='module')
def dataset(library):
    """
    Example dataset.
    """
    return library['ECON']


class TestNamestr:

    def test_decode(self, variable, namestr_bytestring):
        v = xport.v56.Namestr.from_variable(variable, number=1, position=0)
        b = xport.v56.Namestr.from_bytes(namestr_bytestring)
        assert b == v

    def test_encode(self, variable, namestr_bytestring):
        namestr = xport.v56.Namestr.from_variable(variable, number=1, position=0)
        assert bytes(namestr) == namestr_bytestring


class TestMemberHeader:

    def test_decode(self, dataset, member_header_bytestring):
        d = xport.v56.MemberHeader.from_dataset(dataset)
        b = xport.v56.MemberHeader.from_bytes(member_header_bytestring)
        assert b == d

    def test_encode(self, dataset, member_header_bytestring):
        d = xport.v56.MemberHeader.from_dataset(dataset)
        assert bytes(d) == member_header_bytestring


class TestObservations:

    def test_decode(self, dataset, observations_bytestring):
        header = xport.v56.MemberHeader.from_dataset(dataset)
        namestrs = header.namestrs
        obs = xport.v56.Observations.from_bytes(observations_bytestring, namestrs)
        for got, expected in zip(obs, dataset.itertuples(index=False)):
            assert got == expected

    def test_encode(self, dataset, observations_bytestring):
        obs = xport.v56.Observations.from_dataset(dataset)
        i = 0
        for chunk in obs.to_bytes():
            assert chunk == observations_bytestring[i:i + len(chunk)], f'Index {i}'
            i += len(chunk)
        obs = xport.v56.Observations.from_dataset(dataset)
        assert bytes(obs) == observations_bytestring


class TestMember:

    def test_decode(self, dataset, dataset_bytestring):
        member = xport.v56.Member.from_bytes(dataset_bytestring)
        assert (member == dataset).all(axis=None)
        for name in dataset._metadata:
            assert getattr(member, name) == getattr(dataset, name), name
        for k, v in dataset.items():
            u = member[k]
            for name in v._metadata:
                assert getattr(u, name) == getattr(v, name), name

    def test_encode(self, dataset, dataset_bytestring):
        ds = xport.v56.Member(dataset)
        with pytest.warns(UserWarning, match=r'Converting column dtypes'):
            assert bytes(ds) == dataset_bytestring


class TestLibrary:

    def test_decode(self, library, library_bytestring):
        got = xport.v56.Library.from_bytes(library_bytestring)
        assert got == library

    def test_encode(self, library, library_bytestring):
        with pytest.warns(UserWarning, match=r'Converting column dtypes'):
            assert bytes(xport.v56.Library(library)) == library_bytestring

    def test_empty_library(self):
        """
        Verify dumps/loads of an empty library.
        """
        empty = xport.v56.Library()
        bytestring = bytes(empty)
        xport.v56.Library.from_bytes(bytestring)

    def test_empty_member(self):
        """
        Verify dumps/loads of a library with an empty member.
        """
        empty = xport.v56.Library(xport.v56.Member())
        bytestring = bytes(empty)
        xport.v56.Library.from_bytes(bytestring)

    def test_dataframe(self):
        lib = xport.Library(pd.DataFrame({'a': [1]}))
        with pytest.warns(UserWarning, match=r'Converting column dtypes'):
            result = xport.v56.loads(xport.v56.dumps(lib))
        assert (result[''] == lib[None]).all(axis=None)

    def test_no_observations(self):
        """
        Verify dumps/loads of 1 member, 1 variable, 0 observations.
        """
        with pytest.warns(UserWarning, match=r'Set dataset name'):
            library = xport.v56.Library({'x': xport.v56.Member({'a': []})})
        bytestring = bytes(library)
        assert xport.v56.Library.from_bytes(bytestring)


class TestIEEEtoIBM:

    def roundtrip(self, n):
        ibm = xport.v56.ieee_to_ibm(n)
        ieee = xport.v56.ibm_to_ieee(ibm)
        return round(ieee, 9)

    def test_overflow(self):
        with pytest.raises(xport.v56.Overflow):
            xport.v56.ieee_to_ibm(16**63)

    def test_underflow(self):
        with pytest.raises(xport.v56.Underflow):
            xport.v56.ieee_to_ibm(16**-66)

    def test_nan(self):
        n = float('nan')
        assert math.isnan(self.roundtrip(n))

    def test_special_missing_values(self):
        for c in '_' + string.ascii_uppercase:
            n = getattr(xport.NaN, c)
            assert chr(bytes(n)[0]) == c
            assert math.isnan(self.roundtrip(n))

    def test_zero(self):
        assert self.roundtrip(0) == 0

    def test_small_magnitude_integers(self):
        for i in range(-1000, 1000):
            assert self.roundtrip(i) == i

    def test_small_magnitude_floats(self):
        for i in range(-10, 10):
            i /= 1000
            assert self.roundtrip(i) == i

    def test_large_magnitude_floats(self):
        n = int(1e9)
        for i in range(n, n + 100):
            assert self.roundtrip(i) == i

    def test_large_magnitude_floats_with_fraction(self):
        offset = 1e9
        for i in range(100):
            i /= 1e9
            x = i + offset
            assert self.roundtrip(x) == x

    def test_very_small_magnitude_floats(self):
        for i in range(-10, 10):
            i /= 1e6
            assert self.roundtrip(i) == i


class TestEncode:
    """
    Verify various XPORT-encode features.
    """

    def dump_and_load(self, library):
        bytestring = xport.v56.dumps(library)
        return xport.v56.loads(bytestring)

    def test_numeric_type_conversion(self):
        """
        Verify numeric types convert to float when writing.
        """
        numerics = [
            1,
            True,
        ]
        for x in numerics:
            with pytest.warns(UserWarning, match=r'Converting column dtypes'):
                library = xport.Library({'A': xport.Dataset({'x': [x]})})
                output = self.dump_and_load(library)
                assert output['A']['x'].dtype.name == 'float64'
                assert output['A']['x'].iloc[0] == 1.0

    def test_text_null(self):
        # https://github.com/selik/xport/issues/44
        df = pd.DataFrame({
            'a': pd.Series([None], dtype='string'),
            'b': [0],  # Avoid issue #46 by including a numeric column.
        })
        library = self.dump_and_load(df)
        assert list(next(iter(library.values()))['a']) == ['']

    @pytest.mark.skip("Pandas coerces everything to a string")
    def test_invalid_types(self):
        """
        Verify invalid types raise errors on write.
        """
        invalid = [
            b'\x00',
            object(),
            (1, 2, 3),
        ]
        for bad in invalid:
            with pytest.warns(UserWarning, match=r'Converting column dtypes'):
                with pytest.raises(TypeError):
                    library = xport.Library(xport.Dataset({'a': [bad]}))
                    xport.v56.dumps(library)

    def test_invalid_values(self):
        invalid = [
            '\N{snowman}',
        ]
        for bad in invalid:
            library = xport.Library(xport.Dataset({'a': [bad]}))
            with pytest.raises(ValueError):
                with pytest.warns(UserWarning, match=r'Converting column dtypes'):
                    xport.v56.dumps(library)

    def test_dumps_name_and_label_length_validation(self):
        """
        Verify variable and dataset name and label length.
        """
        # Names must be <= 8 characters.
        # Labels must be <= 40 characters.
        # SAS v8 Transport Files allow longer labels.
        invalid = [
            xport.Library(xport.Dataset(), sas_version='a' * 9),
            xport.Library(xport.Dataset(name='a' * 9)),
            xport.Library(xport.Dataset(label='a' * 41)),
            xport.Library(xport.Dataset({'a' * 9: [1.0]})),
            xport.Library(xport.Dataset({'a': xport.Variable([1.0], label='a' * 41)})),
        ]
        for bad_metadata in invalid:
            with pytest.raises(ValueError):
                xport.v56.dumps(bad_metadata)

    def test_troublesome_text(self):
        """
        Some text patterns have been trouble in the past.
        """
        trouble = xport.Variable(["'<>"], dtype='string')
        dataset = xport.Dataset({'a': trouble}, name='trouble')
        library = xport.Library(dataset)
        with pytest.warns(UserWarning, match=r'Converting column dtypes'):
            assert self.dump_and_load(library) == library

    def test_dataset_created(self):
        invalid = datetime(1800, 1, 1)
        ds = xport.Dataset(created=invalid)
        assert ds.created == invalid
        with pytest.raises(ValueError):
            xport.v56.dumps(ds)
        with pytest.raises((TypeError, AttributeError)):
            ds = xport.Dataset(created='2000-Jan-01')
            xport.v56.dumps(ds)

    def test_dataset_modified(self):
        invalid = datetime(2100, 1, 1)
        ds = xport.Dataset(modified=invalid)
        assert ds.modified == invalid
        with pytest.raises(ValueError):
            xport.v56.dumps(ds)
        with pytest.raises((TypeError, AttributeError)):
            ds = xport.Dataset(modified=1)
            xport.v56.dumps(ds)


class TestTextEncoding:
    """
    Validate writing and reading different text encodings.
    """

    def test_default(self):
        """
        Test handling non-ASCII Windows-1252 characters.
        """
        enye = '\u00f1'
        enye.encode('Windows-1252')
        with pytest.raises(UnicodeEncodeError):
            enye.encode('ascii')

        # Metadata defaults to ASCII
        with pytest.raises(UnicodeEncodeError):
            xport.v56.dumps(xport.Dataset(name=enye))
        # TODO: Ensure failure to load a non-ASCII dataset name.

        # Data defaults to Windows-1252
        example = xport.Dataset({'v': xport.Variable([enye])}, name='d')
        bytestring = xport.v56.dumps(example)
        library = xport.v56.loads(bytestring)
        assert library['d']['v'][0] == enye

    def test_ascii(self):
        """
        Test rejecting non-ASCII characters.
        """
        enye = '\u00f1'
        with pytest.raises(UnicodeEncodeError):
            enye.encode('ascii')

        # We're only testing the data, not metadata.
        enye.encode(xport.v56.TEXT_DATA_ENCODING)
        example = xport.Dataset({'v': xport.Variable([enye])})
        with pytest.raises(UnicodeEncodeError):
            with xport.v56._encoding(data='ascii'):
                xport.v56.dumps(example)
        bytestring = xport.v56.dumps(example)
        with pytest.raises(UnicodeDecodeError):
            with xport.v56._encoding(data='ascii'):
                xport.v56.loads(bytestring)

    def test_unicode(self):
        """
        Test encoding and decoding with UTF-8.
        """
        comet = '\u2604'
        with pytest.raises(UnicodeEncodeError):
            comet.encode('ascii')

        # Data
        example = xport.Dataset({'v': xport.Variable([comet])})
        with pytest.raises(UnicodeEncodeError):
            xport.v56.dumps(example)
        with xport.v56._encoding(data='utf-8'):
            bytestring = xport.v56.dumps(example)
            library = xport.v56.loads(bytestring)
        assert library['']['v'][0] == comet

        # Metadata
        example = xport.Dataset(name=comet)
        with xport.v56._encoding(metadata='utf-8'):
            bytestring = xport.v56.dumps(example)
            library = xport.v56.loads(bytestring)
        assert comet in library


class TestEncodeLabels:
    """
    Validate writing data set and variable labels.
    """

    def test_dataset(self):
        """
        Data set label, no variable label.
        """
        example = xport.Dataset(
            {'a': xport.Variable(dtype=object)},
            name='TEST',
            label='This is a test',
        )
        bytestring = xport.v56.dumps(example)
        library = xport.v56.loads(bytestring)
        assert example.label == library['TEST'].label == 'This is a test'

    def test_dataset_deprecated(self):
        """
        The ``Dataset.label`` attribute is now ``Dataset.label``.
        """
        with pytest.warns(DeprecationWarning, match=r'dataset_label'):
            example = xport.Dataset(name='TEST', dataset_label='This is a test')
        bytestring = xport.v56.dumps(example)
        library = xport.v56.loads(bytestring)
        assert example.label == library['TEST'].label == 'This is a test'

    def test_variable(self):
        """
        Only a variable lable, no data set label.
        """
        example = xport.Dataset({'a': xport.Variable(label='b', dtype=object)}, name='TEST')
        bytestring = xport.v56.dumps(example)
        library = xport.v56.loads(bytestring)
        assert example.label is None
        assert library['TEST'].label == ''
        assert example['a'].label == library['TEST']['a'].label == 'b'

    def test_both(self):
        """
        Both data set label and variable label.
        """
        example = xport.Dataset(
            data={'a': xport.Variable(label='b', dtype=object)},
            name='TEST',
            label='Test',
        )
        bytestring = xport.v56.dumps(example)
        library = xport.v56.loads(bytestring)
        assert example.label == library['TEST'].label == 'Test'
        assert example['a'].label == library['TEST']['a'].label == 'b'

    def test_neither(self):
        """
        Neither data set label nor variable label.
        """
        example = xport.Dataset({'a': xport.Variable(dtype=object)}, name='TEST')
        bytestring = xport.v56.dumps(example)
        library = xport.v56.loads(bytestring)
        assert example.label is None
        assert library['TEST'].label == ''
        assert example['a'].label is None
        assert library['TEST']['a'].label == ''



"""
Project metadata, such as the current version number.
"""

# Single-sourcing the package version from the installed package.  This
# has the caveat that the imported version may be different from the
# installed version in some rare circumstances, such as when current
# working directory is the ``src`` folder and the package install did
# not use the ``-e`` or ``--editable`` flag.  Most usage will be from
# the root directory of the project or from outside the project.
# https://packaging.python.org/guides/single-sourcing-package-version/

# Standard Library
import pathlib
from collections import namedtuple

# Community Packages
from pkg_resources import DistributionNotFound, get_distribution

__all__ = [
    '__version__',
]


class Version(namedtuple('Version', 'major minor patch')):
    """
    Version information.
    """

    @classmethod
    def parse(cls, s):
        return Version(*map(int, s.split('.')))

    def __str__(self):
        return '.'.join(map(str, self))


project = pathlib.Path(__file__).parent.name
try:
    __version__ = Version.parse(get_distribution(project).version)
except DistributionNotFound:
    __version__ = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Tests for the core CLI.
"""
# Standard Library
import re
import subprocess
from io import StringIO

# Community Packages
import pandas as pd

# Xport Modules
import xport


def test_help():
    """
    Verify we can show usage help.
    """
    cmd = 'python -m xport --help'
    argv = cmd.split()
    proc = subprocess.run(argv, text=True, capture_output=True)
    assert re.match(r'Usage: xport .*', proc.stdout.splitlines()[0])


def test_version():
    """
    Verify we can show the verison.
    """
    cmd = 'python -m xport --version'
    argv = cmd.split()
    proc = subprocess.run(argv, text=True, capture_output=True)
    line, = proc.stdout.splitlines()
    assert re.match(r'xport, version \S+', line)


def test_executable():
    """
    Verify the command line executable installed correctly.
    """
    executable = 'xport --version'.split()
    imported = 'python -m xport --version'.split()
    proc1 = subprocess.run(executable, text=True, capture_output=True)
    proc2 = subprocess.run(imported, text=True, capture_output=True)
    assert proc1.stdout == proc2.stdout


def test_decode(library, library_bytestring):
    """
    Verify the command line executable can decode a library.
    """
    cmd = 'python -m xport -'
    argv = cmd.split()
    proc = subprocess.run(argv, capture_output=True, input=library_bytestring)
    fp = StringIO(proc.stdout.decode())
    df = pd.read_csv(fp)
    ds = xport.Dataset(df)
    assert (ds == next(iter(library.values()))).all(axis=None)


def test_output_file(library, library_bytestring, tmp_path):
    """
    Verify CLI can write output to a file.
    """
    filepath = tmp_path / 'tmp.csv'
    cmd = f'python -m xport - {filepath}'
    argv = cmd.split()
    subprocess.run(argv, capture_output=True, input=library_bytestring)
    with open(filepath) as f:
        df = pd.read_csv(f)
    ds = xport.Dataset(df)
    assert (ds == next(iter(library.values()))).all(axis=None)



"""
Read and write SAS XPORT/XPT-format files.
"""

# Standard Library
import functools
import json
import logging
import logging.config
import sys

# Community Packages
import click
import yaml

# Xport Modules
import xport
import xport.v56
import xport.v89

__all__ = [
    'cli',
]

try:
    yaml.load = functools.partial(yaml.load, Loader=yaml.CSafeLoader)
except AttributeError:
    yaml.load = functools.partial(yaml.load, Loader=yaml.SafeLoader)

try:
    with open('logging.yml') as file:
        LOG_CONFIG = yaml.load(file)
except FileNotFoundError:
    LOG_CONFIG = {'version': 1}
logging.config.dictConfig(LOG_CONFIG)

LOG = logging.getLogger(__name__)
log_levels = [name for x, name in sorted(logging._levelToName.items()) if x]


@click.command(
    context_settings={'help_option_names': ['-h', '--help']},
)
@click.argument('input', type=click.File('rb'))
@click.argument(
    'output',
    type=click.File('wt'),
    default=sys.stdout,
)
@click.option(
    '--dataset', metavar='NAME', help='Select a dataset by name.  Defaults to the first dataset.'
)
@click.option(
    '--loglevel',
    metavar='LEVEL',
    type=click.Choice(log_levels, case_sensitive=False),
    help=f'Set logging level.  {{{", ".join(log_levels[:-1])}}}',
)
@click.version_option(version=xport.__version__)
def cli(input, output, dataset, loglevel):
    """
    Convert SAS Transport (XPORT) files to comma-separated values (CSV).
    """
    if loglevel:
        for k, config in LOG_CONFIG['loggers'].items():
            config['level'] = loglevel.upper()
        logging.config.dictConfig(LOG_CONFIG)

    LOG.debug('Xport version %s', xport.__version__)
    LOG.debug('CLI arg --loglevel = %r', loglevel)
    LOG.debug('Using logging config %s', json.dumps(LOG_CONFIG, indent=2))

    bytestring = input.read()
    if xport.v89.Library.pattern.match(bytestring):
        library = xport.v89.loads(bytestring)
    else:
        library = xport.v56.loads(bytestring)
    if dataset is not None:
        ds = library[dataset]
    elif library:
        ds = next(iter(library.values()))
    else:
        raise ValueError("Library has no member datasets")
    LOG.info(f'Selected dataset {ds.name!r}')
    ds.to_csv(output, index=False)



"""
Tests for XPT format from SAS versions 8 and 9.
"""

# Standard Library
import datetime

# Community Packages
import pytest

# Xport Modules
import xport.v89


@pytest.fixture()
def library_bytestring():
    """
    Dataset library encoded in SAS Transport Version 8/9 format.
    """
    return b'''\
HEADER RECORD*******LIBV8   HEADER RECORD!!!!!!!000000000000000000000000000000  \
SAS     SAS     SASLIB  6.06    bsd4.2                          11NOV21:22:33:22\
11NOV21:22:33:22                                                                \
HEADER RECORD*******MEMBV8  HEADER RECORD!!!!!!!000000000000000001600000000140  \
HEADER RECORD*******DSCPTV8 HEADER RECORD!!!!!!!000000000000000000000000000000  \
SAS     DATASET                         SASDATA 6.06    bsd4.2  11NOV21:22:33:22\
11NOV21:22:33:22                                                                \
HEADER RECORD*******NAMSTV8 HEADER RECORD!!!!!!!000000000600000000000000000000  \
\x00\x01\x00\x00\x00\x08\x00\x00Float   Floating Point                          \
BEST    \x00\x00\x00\x00\x00\x01\x00\x00BEST    \x00\x00\x00\x00\x00\x00\x00\x00\
Float                           \x00\x0e\
\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x01\x00\x00\x00\x08\x00\x01Double  Double Precision                        \
BEST    \x00\x00\x00\x00\x00\x01\x00\x00BEST    \x00\x00\x00\x00\x00\x00\x00\x08\
Double                          \x00\x10\
\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x01\x00\x00\x00\x08\x00\x02Long    Long Integer                            \
BEST    \x00\x00\x00\x00\x00\x01\x00\x00BEST    \x00\x00\x00\x00\x00\x00\x00\x10\
Long                            \x00\x0c\
\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x01\x00\x00\x00\x08\x00\x03Int     Integer                                 \
BEST    \x00\x00\x00\x00\x00\x01\x00\x00BEST    \x00\x00\x00\x00\x00\x00\x00\x18\
Int                             \x00\x07\
\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x01\x00\x00\x00\x08\x00\x04Byte    Byte. This is a very long label, probabl\
BEST    \x00\x00\x00\x00\x00\x01\x00\x00BEST    \x00\x00\x00\x00\x00\x00\x00 \
Byte                            \x00P\
\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x02\x00\x00\x00\x03\x00\x05Str     String                                  \
$       \x00\x03\x00\x00\x00\x00\x00\x00$       \x00\x03\x00\x00\x00\x00\x00(\
Str                             \x00\x06\
\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\
                                        \
HEADER RECORD*******LABELV8 HEADER RECORD!!!!!!!1                               \
\x00\x04\x00\x04\x00P\
Byte\
Byte. This is a very long label, probably longer than 40 characters, maybe even.\
                                                                      \
HEADER RECORD*******OBSV8   HEADER RECORD!!!!!!!000000000000000000000000000000  \
\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00a  A\x10\x00\x00\x00\x00\x00\x00A\x10\x00\x00\x00\x00\x00\x00A\x10\x00\x00\x00\x00\x00\x00A\x10\x00\x00\x00\x00\x00\x00A\x10\x00\x00\x00\x00\x00\x001  A\x11\x99\x99\xa0\x00\x00\x00A\x11\x99\x99\x99\x99\x99\x9aA \x00\x00\x00\x00\x00\x00A \x00\x00\x00\x00\x00\x00A \x00\x00\x00\x00\x00\x00\xe2\x98\x83.\x00\x00\x00\x00\x00\x00\x00.\x00\x00\x00\x00\x00\x00\x00.\x00\x00\x00\x00\x00\x00\x00.\x00\x00\x00\x00\x00\x00\x00.\x00\x00\x00\x00\x00\x00\x00                                                                       \
'''  # E501: line too long


@pytest.fixture(scope='module')
def library():
    """
    Create a ``Library`` that matches the ``library_bytestring`` fixture.
    """
    ds = xport.Dataset(
        data={
            'Float': [0, 1, 9227469 / 8388608, float('nan')],
            'Double': [0, 1, 1.1, float('nan')],
            'Long': [0, 1, 2, None],
            'Int': [0, 1, 2, None],
            'Byte': [0, 1, 2, None],
            'Str': ['a', '1', '\N{snowman}'.encode().decode('ISO-8859-1'), ''],
        },
        name='DATASET',
        label='',
        dataset_type='',
    )
    ds.created = ds.modified = datetime.datetime(2021, 11, 11, 22, 33, 22)
    ds.sas_os = 'bsd4.2'
    ds.sas_version = '6.06'
    for name, variable in ds.items():
        variable.width = 8
        variable.format = xport.Format.from_spec('BEST0.', justify=xport.FormatAlignment.RIGHT)
        variable.informat = 'BEST0.'
    ds['Float'].label = 'Floating Point'
    ds['Double'].label = 'Double Precision'
    ds['Long'].label = 'Long Integer'
    ds['Int'].label = 'Integer'
    ds['Byte'].label = (
        'Byte. This is a very long label, probably longer than 40 characters, maybe even.'
    )
    ds['Str'].label = 'String'
    ds['Str'].width = 3
    ds['Str'].vtype = xport.VariableType.CHARACTER
    ds['Str'].format = '$3.'
    ds['Str'].informat = '$3.'
    return xport.Library(
        members=[ds],
        created=ds.created,
        modified=ds.modified,
        sas_os=ds.sas_os,
        sas_version=ds.sas_version,
    )


class TestLibrary:
    """
    Test SAS Transport v8/9 features.
    """

    def test_decode_labels(self, library, library_bytestring):
        """Test decoding long variable names and labels."""
        got = xport.v89.loads(library_bytestring)
        assert len(got['DATASET']['Byte'].label) > 40
        assert got['DATASET']['Byte'].label == library['DATASET']['Byte'].label
        assert ((got['DATASET'] == library['DATASET'])
                | (got['DATASET'].isna() & library['DATASET'].isna())).all(axis=None)

    def test_encode_labels(self, library, library_bytestring):
        """Test encoding long variable names and labels."""
        got = xport.v89.dumps(library)
        assert got == library_bytestring

    @pytest.mark.skip("We need verified example data.")
    def test_decode_formats(self):
        """Test decoding long format descriptions."""
        assert False

    @pytest.mark.skip("We need verified example data.")
    def test_encode_formats(self):
        """Test encoding long format descriptions."""
        assert False


class TestTextEncoding:
    """
    Validate writing and reading different text encodings.
    """

    def test_default(self):
        """
        Test handling non-ASCII Windows-1252 characters.
        """
        enye = '\u00f1'
        enye.encode('Windows-1252')
        with pytest.raises(UnicodeEncodeError):
            enye.encode('ascii')

        # Metadata defaults to ASCII
        with pytest.raises(UnicodeEncodeError):
            xport.v89.dumps(xport.Dataset(name=enye))
        # TODO: Ensure failure to load a non-ASCII dataset name.

        # Data defaults to Windows-1252
        example = xport.Dataset({'v': xport.Variable([enye])}, name='d')
        bytestring = xport.v89.dumps(example)
        library = xport.v89.loads(bytestring)
        assert library['d']['v'][0] == enye

    def test_ascii(self):
        """
        Test rejecting non-ASCII characters.
        """
        enye = '\u00f1'
        with pytest.raises(UnicodeEncodeError):
            enye.encode('ascii')

        # We're only testing the data, not metadata.
        example = xport.Dataset({'v': xport.Variable([enye])})
        with pytest.raises(UnicodeEncodeError):
            with xport.v89._encoding(data='ascii'):
                xport.v89.dumps(example)
        bytestring = xport.v89.dumps(example)
        with pytest.raises(UnicodeDecodeError):
            with xport.v89._encoding(data='ascii'):
                xport.v89.loads(bytestring)

    def test_unicode(self):
        """
        Test encoding and decoding with UTF-8.
        """
        comet = '\u2604'
        with pytest.raises(UnicodeEncodeError):
            comet.encode('ascii')

        # Data
        example = xport.Dataset({'v': xport.Variable([comet])})
        with pytest.raises(UnicodeEncodeError):
            xport.v89.dumps(example)
        with xport.v89._encoding(data='utf-8'):
            bytestring = xport.v89.dumps(example)
            library = xport.v89.loads(bytestring)
        assert library['']['v'][0] == comet

        # Metadata
        example = xport.Dataset(name=comet)
        with xport.v89._encoding(metadata='utf-8'):
            bytestring = xport.v89.dumps(example)
            library = xport.v89.loads(bytestring)
        assert comet in library


class TestEncodeLabels:
    """
    Validate writing data set and variable labels.
    """

    def test_dataset(self):
        """
        Data set label, no variable label.
        """
        example = xport.Dataset({'a': xport.Variable()}, name='TEST', label='This is a test')
        bytestring = xport.v89.dumps(example)
        library = xport.v89.loads(bytestring)
        assert example.label == library['TEST'].label == 'This is a test'

    def test_dataset_deprecated(self):
        """
        The ``Dataset.label`` attribute is now ``Dataset.label``.
        """
        example = xport.Dataset(name='TEST', dataset_label='This is a test')
        bytestring = xport.v89.dumps(example)
        library = xport.v89.loads(bytestring)
        assert example.label == library['TEST'].label == 'This is a test'

    def test_variable(self):
        """
        Only a variable lable, no data set label.
        """
        example = xport.Dataset({'a': xport.Variable(label='b')}, name='TEST')
        bytestring = xport.v89.dumps(example)
        library = xport.v89.loads(bytestring)
        assert example.label is None
        assert library['TEST'].label == ''
        assert example['a'].label == library['TEST']['a'].label == 'b'

    def test_both(self):
        """
        Both data set label and variable label.
        """
        example = xport.Dataset(
            data={'a': xport.Variable(label='b')},
            name='TEST',
            label='Test',
        )
        bytestring = xport.v89.dumps(example)
        library = xport.v89.loads(bytestring)
        assert example.label == library['TEST'].label == 'Test'
        assert example['a'].label == library['TEST']['a'].label == 'b'

    def test_neither(self):
        """
        Neither data set label nor variable label.
        """
        example = xport.Dataset({'a': xport.Variable()}, name='TEST')
        bytestring = xport.v89.dumps(example)
        library = xport.v89.loads(bytestring)
        assert example.label is None
        assert library['TEST'].label == ''
        assert example['a'].label is None
        assert library['TEST']['a'].label == ''



"""
Read and write the SAS XPORT/XPT file format from SAS Version 5 or 6.

The SAS V5 Transport File format, also called XPORT, or simply XPT, ...
"""

# All "records" are 80 bytes long, padded if necessary.
# Character data are ASCII-encoded.
# Integer data are IBM-style integer format.
# Floating point data are IBM-style double format.

# Standard Library
import contextlib
import logging
import math
import re
import struct
import warnings
from collections.abc import Iterator, Mapping
from datetime import datetime

# Community Packages
import pandas as pd

# Xport Modules
import xport

__all__ = [
    'load',
    'loads',
    'dump',
    'dumps',
    '_encoding',
]

LOG = logging.getLogger(__name__)

# TODO: Make text encoding a method argument, not a global.
# TODO: Consider using Windows-1252 as the default, like HTML5.
TEXT_DATA_ENCODING = 'ISO-8859-1'
TEXT_METADATA_ENCODING = 'ascii'


@contextlib.contextmanager
def _encoding(data=TEXT_DATA_ENCODING, metadata=TEXT_METADATA_ENCODING):
    """
    Temporarily change the module's text encoding.
    """
    global TEXT_DATA_ENCODING
    global TEXT_METADATA_ENCODING
    stash = {'data': TEXT_DATA_ENCODING, 'metadata': TEXT_METADATA_ENCODING}
    try:
        TEXT_DATA_ENCODING = data
        TEXT_METADATA_ENCODING = metadata
        yield
    finally:
        LOG.debug(f'Reverting text encoding to {stash}')
        TEXT_DATA_ENCODING = stash['data']
        TEXT_METADATA_ENCODING = stash['metadata']


class Overflow(ArithmeticError):
    """Number too large to express."""


class Underflow(ArithmeticError):
    """Number too small to express, rounds to zero."""


class Namestr:
    """
    Variable metadata from a SAS Version 5 or 6 Transport (XPORT) file.
    """

    # Here is the C structure definition for the namestr record:
    #
    # struct NAMESTR {
    #    short ntype;       /* VARIABLE TYPE: 1=NUMERIC, 2=CHAR       */
    #    short nhfun;       /* HASH OF NNAME (always 0)               */
    #    short nlng;        /* LENGTH OF VARIABLE IN OBSERVATION      */
    #    short nvar0;       /* VARNUM                                 */
    #    char8 nname;       /* NAME OF VARIABLE                       */
    #    char40 nlabel;     /* LABEL OF VARIABLE                      */
    #    char8 nform;       /* NAME OF FORMAT                         */
    #    short nfl;          /* FORMAT FIELD LENGTH OR 0               */
    #    short nfd;         /* FORMAT NUMBER OF DECIMALS              */
    #    short nfj;         /* 0=LEFT JUSTIFICATION, 1=RIGHT JUST     */
    #    char nfill[2];      /* (UNUSED, FOR ALIGNMENT AND FUTURE)     */
    #    char8 niform;      /* NAME OF INPUT FORMAT                   */
    #    short nifl;         /* INFORMAT LENGTH ATTRIBUTE              */
    #    short nifd;        /* INFORMAT NUMBER OF DECIMALS            */
    #    long npos;         /* POSITION OF VALUE IN OBSERVATION       */
    #    char rest[52];     /* remaining fields are irrelevant         */
    #    };
    #
    # Note that the length given in the last 4 bytes of the member
    # header record indicates the actual number of bytes for the NAMESTR
    # structure. The size of the structure listed above is 140 bytes.
    # Under VAX/VMS, the size will be 136 bytes, meaning that the 'rest'
    # variable may be truncated.

    fmts = {
        140: '>hhhh8s40s8shhh2s8shhl52s',
        136: '>hhhh8s40s8shhh2s8shhl48s',
    }

    def __init__(self, vtype, length, number, name, label, format, informat, position):
        """
        Initialize a ``Namestr``.
        """
        self.vtype = vtype
        self.length = length
        self.number = number
        self.name = name
        self.label = label
        self.format = format
        self.informat = informat
        self.position = position

    def __eq__(self, other):
        """Equality."""
        attributes = [
            'vtype',
            'name',
            'label',
            'format',
            'informat',
            'position',
        ]
        return all(getattr(self, name) == getattr(other, name) for name in attributes)

    @classmethod
    def from_variable(cls, variable: xport.Variable, number=None, position=None):
        """
        Construct a ``Namestr`` from an ``xport.Variable``.
        """
        if variable.vtype is not None:
            vtype = variable.vtype
        elif variable.dtype.kind in {'f', 'i'}:
            vtype = xport.VariableType.NUMERIC
        elif variable.dtype.kind == 'O':
            vtype = xport.VariableType.CHARACTER
        elif variable.dtype.kind == 'b':
            # We'll encode Boolean columns as 1 if True else 0.
            vtype = xport.VariableType.NUMERIC
        else:
            raise TypeError(f'{type(variable).__name__}.dtype {variable.dtype} not supported')

        if variable.width is not None:
            length = variable.width
        elif vtype == xport.VariableType.NUMERIC:
            length = 8
        else:
            # TODO: Avoid encoding twice, once here and once in ``Observations``.
            length = variable.str.encode(TEXT_DATA_ENCODING).str.len().max()
        try:
            length = max(1, length)  # We need at least 1 byte per value.
        except TypeError:
            length = 1
        return cls(
            vtype=vtype,
            length=length,
            number=number,
            name=variable.name,
            label=variable.label,
            format=variable.format if variable.format is not None else xport.Format(),
            informat=variable.informat if variable.informat is not None else xport.Informat(),
            position=position,
        )

    @classmethod
    def from_bytes(cls, bytestring: bytes):
        """
        Construct a ``Namestr`` from an XPORT-format byte string.
        """
        LOG.debug(f'Decode {type(cls).__name__}')
        # dtype='float' if vtype == xport.VariableType.NUMERIC else 'string'
        size = len(bytestring)
        if size == 136:
            warnings.warn('File written on VAX/VMS, module behavior not tested')
        fmt = cls.fmts[size]
        tokens = struct.unpack(fmt, bytestring)
        return cls(
            vtype=xport.VariableType(tokens[0]),
            length=tokens[2],
            number=tokens[3],
            name=tokens[4].strip(b'\x00').decode(TEXT_METADATA_ENCODING).rstrip(),
            label=tokens[5].strip(b'\x00').decode(TEXT_METADATA_ENCODING).rstrip(),
            format=xport.Format.from_struct_tokens(*tokens[6:10]),
            informat=xport.Informat.from_struct_tokens(*tokens[11:14]),
            position=tokens[14],
        )

    def __bytes__(self):
        """
        Encode in XPORT-format.
        """
        LOG.debug(f'Encode {type(self).__name__}')
        fmt = self.fmts[140]
        format_name = self.format.name.encode('ascii')
        if len(format_name) > 8:
            raise ValueError(f'ASCII-encoded format name {format_name} longer than 8 characters')
        informat_name = self.informat.name.encode('ascii')
        if len(informat_name) > 8:
            raise ValueError(f'ASCII-encoded format name {informat_name} longer than 8 characters')
        if self.number is None:
            raise ValueError('Variable number not assigned')
        if self.position is None:
            raise ValueError('Variable position not assigned')
        return struct.pack(
            fmt,
            self.vtype,
            0,  # "Hash" of name, always 0.
            self.length,
            self.number,
            text_encode(self, 'name', 8),
            text_encode(self, 'label', 40),
            format_name.ljust(8),
            self.format.length,
            self.format.decimals,
            self.format.justify,
            b'',  # Unused
            informat_name.ljust(8),
            self.informat.length,
            self.informat.decimals,
            self.position,
            b'',  # Padding
        )


class MemberHeader(Mapping):
    """
    Dataset metadata from a SAS Version 5 or 6 Transport (XPORT) file.
    """

    # 4. Member header records
    #    Both of these records occur for every member in the file.
    #
    #    HEADER RECORD*******MEMBER HEADER RECORD!!!!!!!
    #    000000000000000001600000000140
    #    HEADER RECORD*******DSCRPTR HEADER RECORD!!!!!!!
    #    000000000000000000000000000000
    #
    #    Note the 0140 that appears in the member header record above.
    #    This value specifies the size of the variable descriptor
    #    (NAMESTR) record that is described later in this document. On
    #    the VAX/VMS operating system, the value will be 0136 instead of
    #    0140. This means that the descriptor will be only 136 bytes
    #    instead of 140.
    #
    # 5. Member header data ... as C structure:
    #
    #       struct REAL_HEADER {
    #          char sas_symbol[8];      /* "SAS"                    */
    #          char sas_dsname[8];      /* dataset name             */
    #          char sasdata[8];         /* "SASDATA"                */
    #          char sasver[8];          /* version of SAS used      */
    #          char sas_osname[8];      /* operating system used    */
    #          char blanks[24];
    #          char sas_create[16];     /* datetime created         */
    #          };
    #
    #    The second header record as C structure:
    #
    #       struct SECOND_HEADER {
    #          char dtmod[16];            /* date modified           */
    #          char padding[16];
    #          char dslabel[40];          /* dataset label          */
    #          char dstype[8]             /* dataset type           */
    #          };
    #
    # 6. Namestr header record
    #    One for each member
    #
    #       HEADER RECORD*******NAMESTR HEADER RECORD!!!!!!!
    #       000000xxxx0000000000000 0000000
    #
    #    In this header record, xxxx is the number of variables in the
    #    data set, displayed with blank-padded numeric characters. For
    #    example, for 2 variables, xxxx=0002.
    #
    # 7. Namestr records
    #    Each namestr field is 140 bytes long, but the fields are
    #    streamed together and broken in 80-byte pieces. If the last
    #    byte of the last namestr field does not fall in the last byte
    #    of the 80-byte record, the record is padded with ASCII blanks
    #    to 80 bytes.
    #
    # 8. Observation header
    #
    #       HEADER RECORD*******OBS HEADER RECORD!!!!!!!
    #       000000000000000000000000 000000

    pattern = re.compile(
        # Header line 1
        rb'HEADER RECORD\*{7}MEMBER  HEADER RECORD\!{7}0{17}'
        rb'160{8}(?P<descriptor_size>140|136)  '
        # Header line 2
        rb'HEADER RECORD\*{7}DSCRPTR HEADER RECORD\!{7}0{30} {2}'
        # Header line 3
        rb'SAS {5}(?P<name>.{8})SASDATA '
        rb'(?P<version>.{8})(?P<os>.{8}) {24}(?P<created>.{16})'
        # Header line 4
        rb'(?P<modified>.{16}) {16}'
        rb'(?P<label>.{40})(?P<type>    DATA|    VIEW| {8})'
        # Namestrs
        rb'HEADER RECORD\*{7}NAMESTR HEADER RECORD\!{7}0{6}'
        rb'(?P<n_variables>.{4})0{20} {2}'
        rb'(?P<namestrs>.*?)'
        rb'HEADER RECORD\*{7}OBS {5}HEADER RECORD\!{7}0{30} {2}',
        # Observations ... until EOF or another Member.
        re.DOTALL,
    )

    def __init__(
        self,
        name,
        dataset_label,
        dataset_type,
        created,
        modified,
        sas_os,
        sas_version,
        namestrs=(),
    ):
        """
        Initialize a ``MemberHeader``.
        """
        self.name = name
        self.dataset_label = dataset_label
        self.dataset_type = dataset_type
        self.created = created
        self.modified = modified
        self.sas_os = sas_os
        self.sas_version = sas_version
        self.namestrs = {ns.name: ns for ns in namestrs}

    def __repr__(self):
        """
        Format for the REPL.
        """
        metadata = (name.strip('_') for name in xport.Dataset._metadata)
        metadata = {name: getattr(self, name) for name in metadata}
        metadata = (f'{k.title()}: {v}' for k, v in metadata.items() if v is not None)
        return f'<{type(self).__name__} {", ".join(metadata)}>'

    def __getitem__(self, name):
        """
        Get a namestr by variable name.
        """
        return self.namestrs[name]

    def __iter__(self):
        """
        Get an iterator of variable names.
        """
        return iter(self.namestrs)

    def __len__(self):
        """
        Get the number of variables/namestrs.
        """
        return len(self.namestrs)

    def __eq__(self, other):
        """
        Check equality.
        """
        if not isinstance(other, MemberHeader):
            raise TypeError(f"Can't compare {type(self).__name__} with {type(other).__name__}")
        metadata = (name.strip('_') for name in xport.Dataset._metadata)
        return super().__eq__(other) and all(
            getattr(self, name) == getattr(other, name) for name in metadata
        )

    @classmethod
    def from_dataset(cls, dataset: xport.Dataset, Namestr=Namestr, variable_enumeration_start=1):
        """
        Construct a ``MemberHeader`` from an ``xport.Dataset``.
        """
        namestrs = []
        p = 0
        for i, (k, v) in enumerate(dataset.items(), variable_enumeration_start):
            ns = Namestr.from_variable(v, number=i)
            ns.position = p
            p += ns.length
            namestrs.append(ns)
        return cls(
            name=dataset.name,
            dataset_label=dataset.dataset_label,
            dataset_type=dataset.dataset_type,
            created=dataset.created,
            modified=dataset.modified,
            sas_os=dataset.sas_os,
            sas_version=dataset.sas_version,
            namestrs=namestrs,
        )

    @classmethod
    def from_bytes(cls, bytestring: bytes, Namestr=Namestr):
        """
        Construct a ``MemberHeader`` from an XPORT-format byte string.
        """
        LOG.debug(f'Decode {type(cls).__name__}')
        mo = cls.pattern.search(bytestring)  # TODO: Why ``search``, not ``match``?
        if mo is None:
            raise ValueError('No member header found')

        def chunks():
            stride = int(mo['descriptor_size'])
            for i in range(0, len(mo['namestrs']), stride):
                chunk = mo['namestrs'][i:i + stride]
                if len(chunk) == stride:
                    yield chunk

        n = int(mo['n_variables'])
        namestrs = [Namestr.from_bytes(b) for b in chunks()]
        if len(namestrs) != n:
            raise ValueError(f'Expected {n}, got {len(namestrs)}')
        self = cls(
            name=mo['name'].strip(b'\x00').decode(TEXT_METADATA_ENCODING).strip(),
            dataset_label=mo['label'].strip(b'\x00').decode(TEXT_METADATA_ENCODING).strip(),
            dataset_type=mo['type'].strip(b'\x00').decode(TEXT_METADATA_ENCODING).strip(),
            sas_os=mo['os'].strip(b'\x00').decode(TEXT_METADATA_ENCODING).strip(),
            sas_version=mo['version'].strip().decode(TEXT_METADATA_ENCODING),
            created=strptime(mo['created']),
            modified=strptime(mo['modified']),
            namestrs=namestrs,
        )
        return self

    template = f'''\
HEADER RECORD{'*' * 7}MEMBER  HEADER RECORD{'!' * 7}{'0' * 17}16{'0' * 8}140  \
HEADER RECORD{'*' * 7}DSCRPTR HEADER RECORD{'!' * 7}{'0' * 30}  \
SAS     %(name)8bSASDATA %(version)8b%(os)8b{' ' * 24}%(created)16b\
%(modified)16b{' ' * 16}%(label)40b%(type)8b\
HEADER RECORD{'*' * 7}NAMESTR HEADER RECORD{'!' * 7}{'0' * 6}\
%(n_variables)04d{'0' * 20}  \
%(namestrs)b\
HEADER RECORD{'*' * 7}OBS     HEADER RECORD{'!' * 7}{'0' * 30}  \
'''.encode('ascii')

    def __bytes__(self):
        """
        Encode in XPORT format.
        """
        LOG.debug(f'Encode {type(self).__name__}')
        namestrs = b''.join(bytes(ns) for ns in self.values())
        if len(namestrs) % 80:
            namestrs += b' ' * (80 - len(namestrs) % 80)
        return self.template % {
            b'name': text_encode(self, 'name', 8),
            b'label': text_encode(self, 'dataset_label', 40),
            b'type': text_encode(self, 'dataset_type', 8),
            b'n_variables': len(self),
            b'os': text_encode(self, 'sas_os', 8),
            b'version': text_encode(self, 'sas_version', 8),
            b'created': strftime(self.created if self.created else datetime.now()),
            b'modified': strftime(self.modified if self.modified else datetime.now()),
            b'namestrs': namestrs,
        }


class Observations(Iterator):
    """
    Data from a SAS Version 5 or 6 Transport (XPORT) file.

    ``Observations`` is an iterator, yielding observations as tuples.
    """

    # 9. Data records
    #    Data records are streamed in the same way that namestrs are.
    #    There is ASCII blank padding at the end of the last record if
    #    necessary. There is no special trailing record.

    def __init__(self, observations, header=None):
        """
        Initialize from an iterable of observations.
        """
        self.it = iter(observations)
        self.header = header

    def __next__(self):
        """
        Get the next item from the iterator.
        """
        return next(self.it)

    @classmethod
    def from_dataset(cls, dataset):
        """
        Yield observations from an ``xport.Dataset``.
        """
        return cls(
            observations=dataset.itertuples(index=False, name='Observation'),
            header=MemberHeader.from_dataset(dataset),
        )

    @classmethod
    def from_bytes(cls, bytestring, header):
        """
        Yield observations from an XPORT-format byte string.
        """
        LOG.debug(f'Decode {type(cls).__name__}')

        def character_decode(s):
            return s.decode(TEXT_DATA_ENCODING).rstrip()

        converters = []
        for namestr in header.values():
            if namestr.vtype == xport.VariableType.NUMERIC:
                converters.append(ibm_to_ieee)
            else:
                converters.append(character_decode)

        def iterator():
            sizes = [namestr.length for namestr in header.values()]
            fmt = ''.join(f'{x}s' for x in sizes)
            stride = sum(sizes)
            if stride == 0:
                return
            # TODO: The SAS Transport v5 specification says the sentinel
            #       character is b' ', but people report b'\x00' is used
            #       in some files.  Unfortunately, that would make rows
            #       with all zeros indistiguishable from the sentinel.
            sentinel = b' ' * stride
            mview = memoryview(bytestring)
            for i in range(0, len(mview), stride):
                chunk = mview[i:i + stride]
                if len(chunk) != stride or chunk == sentinel:
                    return
                # TODO: If only characters and the last row is all empty
                #       or spaces, it's indistinguishable from padding.
                #       https://github.com/selik/xport/issues/46
                tokens = struct.unpack(fmt, chunk)
                yield tuple(f(v) for f, v in zip(converters, tokens))

        return cls(iterator(), header)

    def to_bytes(self):
        """
        Get an iterator of XPORT-encoded observations.
        """

        def character_encoder(length):

            def encoder(s):
                try:
                    return s.encode(TEXT_DATA_ENCODING).ljust(length)
                except AttributeError:
                    return b' '
                # If handling errors from None, NAType, etc. is a
                # bottleneck, we should ``fillna`` before creating the
                # values iterator.

            return encoder

        fmt = ''.join(f'{namestr.length}s' for namestr in self.header.values())
        converters = []
        for namestr in self.header.values():
            if namestr.vtype == xport.VariableType.NUMERIC:
                converters.append(ieee_to_ibm)
            else:
                converters.append(character_encoder(namestr.length))
        for t in self:
            g = (f(v) for f, v in zip(converters, t))
            yield struct.pack(fmt, *g)

    def __bytes__(self):
        """
        Encode in XPORT-format.
        """
        LOG.debug(f'Encode {type(self).__name__}')
        observations = b''.join(self.to_bytes())
        if len(observations) % 80:
            observations += b' ' * (80 - len(observations) % 80)
        return observations


class Member(xport.Dataset):
    """
    Dataset from a SAS Version 5 or 6 Transport (XPORT) file.
    """

    @classmethod
    def from_header(cls, header):
        """
        Create an empty ``Member`` with metadata from a ``MemberHeader``.
        """
        variables = {
            namestr.name: xport.Variable(
                dtype='float' if namestr.vtype == xport.VariableType.NUMERIC else 'string',
                name=namestr.name,
                label=namestr.label,
                vtype=namestr.vtype,
                width=namestr.length,
                format=namestr.format,
                informat=namestr.informat,
            )
            for namestr in header.values()
        }
        public = (name.lstrip('_') for name in cls._metadata)
        self = cls(variables, **{name: getattr(header, name) for name in public})
        return self

    @classmethod
    def from_bytes(cls, bytestring, MemberHeader=MemberHeader, Observations=Observations):
        """
        Decode the first ``Member`` from an XPORT-format byte string.
        """
        LOG.debug(f'Decode {type(cls).__name__}')
        mview = memoryview(bytestring)
        matches = MemberHeader.pattern.finditer(mview)

        try:
            mo = next(matches)
        except StopIteration:
            raise ValueError('No member header found')
        i = mo.end(0)

        try:
            mo = next(matches)
        except StopIteration:
            j = None
        else:
            j = mo.start(0)

        header = MemberHeader.from_bytes(mview[:i])
        observations = Observations.from_bytes(mview[i:j], header)

        # This awkwardness works around Pandas subclasses misbehaving.
        # ``DataFrame.append`` discards subclass attributes.  Lame.
        head = cls.from_header(header)
        data = Member(pd.DataFrame.from_records(observations, columns=list(header)))
        data.copy_metadata(head)
        LOG.info(f'Decoded XPORT dataset {data.name!r}')
        LOG.debug('%s', data)
        return data

    def __bytes__(self):
        """
        Encode in XPORT-format.
        """
        return self._bytes()

    def _bytes(self, MemberHeader=MemberHeader, Observations=Observations):
        LOG.debug(f'Encode {type(self).__name__}')
        dtype_kind_conversions = {
            'O': 'string',
            'b': 'float',
            'i': 'float',
        }
        dtypes = self.dtypes.to_dict()
        conversions = {}
        for column, dtype in dtypes.items():
            try:
                conversions[column] = dtype_kind_conversions[dtype.kind]
            except KeyError:
                continue
        if conversions:
            warnings.warn(f'Converting column dtypes {conversions}')
            # BUG: ``DataFrame.copy`` mutates and discards ``Variable`` metadata.
            # self = self.copy()  # Don't mutate!
            cpy = xport.Dataset({k: v for k, v in self.items()})
            for column, dtype in conversions.items():
                LOG.warning(f'Converting column {column!r} from {dtypes[column]} to {dtype}')
                try:
                    cpy[column] = cpy[column].astype(dtype)
                except Exception:
                    raise TypeError(f'Could not coerce column {column!r} to {dtype}')
            cpy.copy_metadata(self)
            self = cpy
        header = bytes(MemberHeader.from_dataset(self))
        observations = bytes(Observations.from_dataset(self))
        return header + observations


class Library(xport.Library):
    """
    Collection of datasets from a SAS Version 5 or 6 Transport file.
    """

    # 1. The first header record:
    #
    #   HEADER RECORD*******LIBRARY HEADER RECORD!!!!!!!
    #   000000000000000000000000000000
    #
    # 2. The first real header record ... as a C structure:
    #
    #   struct REAL_HEADER {
    #      char sas_symbol[2][8];       /* "SAS", twice             */
    #      char saslib[8];              /* "SASLIB"                 */
    #      char sasver[8];              /* version of SAS used      */
    #      char sas_os[8];              /* operating system used    */
    #      char blanks[24];
    #      char sas_create[16];         /* datetime created         */
    #      };
    #
    # 3. Second real header record
    #
    #       ddMMMyy:hh:mm:ss
    #
    #    In this record, the string is the datetime modified. Most
    #    often, the datetime created and datetime modified will always
    #    be the same. Pad with ASCII blanks to 80 bytes. Note that only
    #    a 2-digit year appears. If any program needs to read in this
    #    2-digit year, be prepared to deal with dates in the 1900s or
    #    the 2000s.

    pattern = re.compile(
        rb'HEADER RECORD\*{7}LIBRARY HEADER RECORD\!{7}0{30} {2}'
        rb'SAS {5}SAS {5}SASLIB {2}'
        rb'(?P<version>.{8})(?P<os>.{8}) {24}(?P<created>.{16})'
        rb'(?P<modified>.{16}) {64}'
        rb'(?P<members>.*)',
        re.DOTALL,
    )

    @classmethod
    def from_bytes(cls, bytestring, MemberHeader=MemberHeader, Member=Member):
        """
        Parse a SAS XPORT document from a byte string.
        """
        LOG.debug(f'Decoding {cls.__name__} from {len(bytestring)} bytes')
        mview = memoryview(bytestring)
        mo = cls.pattern.match(mview)
        if mo is None:
            lines = [mview[i * 80:(i + 1) * 80].tobytes() for i in range(8)]
            LOG.error('Document begins with' + '\n%s' * len(lines), *lines)
            raise ValueError('Document does not match SAS Version 5 or 6 Transport (XPORT) format')

        matches = MemberHeader.pattern.finditer(mview)
        indices = [m.start(0) for m in matches] + [None]
        chunks = (mview[i:j] for i, j in zip(indices, indices[1:]))
        self = cls(
            members=map(Member.from_bytes, chunks),
            created=strptime(mo['created']),
            modified=strptime(mo['modified']),
            sas_os=mo['os'].strip(b'\x00').decode(TEXT_METADATA_ENCODING).strip(),
            sas_version=mo['version'].strip(b'\x00').decode(TEXT_METADATA_ENCODING).strip(),
        )
        LOG.info(f'Decoded {self}')
        return self

    template = f'''\
HEADER RECORD{'*' * 7}LIBRARY HEADER RECORD{'!' * 7}{'0' * 30}  \
SAS     SAS     SASLIB  \
%(version)8b%(os)8b{' ' * 24}%(created)16b\
%(modified)16b{' ' * 64}\
%(members)b\
'''.encode('ascii')

    def __bytes__(self):
        """
        XPORT-format bytes string.
        """
        return self._bytes()

    def _bytes(self, Member=Member):
        return self.template % {
            b'version': text_encode(self, 'sas_version', 8),
            b'os': text_encode(self, 'sas_os', 8),
            b'created': strftime(self.created if self.created else datetime.now()),
            b'modified': strftime(self.modified if self.modified else datetime.now()),
            b'members': b''.join(bytes(Member(member)) for member in self.values()),
        }


def text_encode(obj, name, n):
    """
    Encode and check resulting byte string length.
    """
    # The attribute name is a parameter to provide a useful error message.
    value = getattr(obj, name)
    if value is None:
        value = ''
    bytestring = value.encode(TEXT_METADATA_ENCODING).ljust(n)
    if len(bytestring) > n:
        raise ValueError(f'Encoded {name} {bytestring} longer than {n} characters')
    return bytestring


def strptime(timestring):
    """
    Parse a datetime from an XPT format string.

    All text in an XPT document are ASCII-encoded.  This function
    expects a bytes string in the "ddMMMyy:hh:mm:ss" format.  For
    example, ``b'16FEB11:10:07:55'``.  Note that XPT supports only
    2-digit years, which are expected to be either 1900s or 2000s.
    """
    text = timestring.decode('ascii')
    return datetime.strptime(text, '%d%b%y:%H:%M:%S')


def strftime(dt, minimum=datetime(1900, 1, 1), maximum=datetime(2100, 1, 1)):
    """
    Convert a datetime to an XPT format byte string.
    """
    if dt < minimum:
        raise ValueError(f'y2k never left! {dt} is in year {dt:%y}')
    if dt >= maximum:
        raise ValueError(f'y2k never left! {dt} is in year {dt:%y}')
    return dt.strftime('%d%b%y:%H:%M:%S').upper().encode('ascii')


def ibm_to_ieee(ibm: bytes) -> float:
    """
    Convert IBM-format floating point (bytes) to IEEE 754 64-bit (float).
    """
    # IBM mainframe:    sign * 0.mantissa * 16 ** (exponent - 64)
    # Python uses IEEE: sign * 1.mantissa * 2 ** (exponent - 1023)

    # Pad-out to 8 bytes if necessary. We expect 2 to 8 bytes, but
    # there's no need to check; bizarre sizes will cause a struct
    # module unpack error.
    ibm = ibm.ljust(8, b'\x00')

    # parse the 64 bits of IBM float as one 8-byte unsigned long long
    ulong, = struct.unpack('>Q', ibm)

    # IBM: 1-bit sign, 7-bits exponent, 56-bits mantissa
    sign = ulong & 0x8000000000000000
    exponent = (ulong & 0x7f00000000000000) >> 56
    mantissa = ulong & 0x00ffffffffffffff

    if mantissa == 0:
        if ibm[:1] == b'\x00':
            return 0.0
        elif ibm[:1] == b'\x80':
            return -0.0
        elif ibm[:1] == b'.':
            return float('nan')
        elif ibm[:1] in b'_ABCDEFGHIJKLMNOPQRSTUVWXYZ':
            return getattr(xport.NaN, chr(ibm[0]))
        else:
            raise ValueError('Neither "true" zero nor NaN: %r' % ibm)

    # IBM-format exponent is base 16, so the mantissa can have up to 3
    # leading zero-bits in the binary mantissa. IEEE format exponent
    # is base 2, so we don't need any leading zero-bits and will shift
    # accordingly. This is one of the criticisms of IBM-format, its
    # wobbling precision.
    if ulong & 0x0080000000000000:
        shift = 3
    elif ulong & 0x0040000000000000:
        shift = 2
    elif ulong & 0x0020000000000000:
        shift = 1
    else:
        shift = 0
    mantissa >>= shift

    # clear the 1 bit to the left of the binary point
    # this is implicit in IEEE specification
    mantissa &= 0xffefffffffffffff

    # IBM exponent is excess 64, but we subtract 65, because of the
    # implicit 1 left of the radix point for the IEEE mantissa
    exponent -= 65
    # IBM exponent is base 16, IEEE is base 2, so we multiply by 4
    exponent <<= 2
    # IEEE exponent is excess 1023, but we also increment for each
    # right-shift when aligning the mantissa's first 1-bit
    exponent += shift + 1023

    # IEEE: 1-bit sign, 11-bits exponent, 52-bits mantissa
    # We didn't shift the sign bit, so it's already in the right spot
    ieee = sign | (exponent << 52) | mantissa
    return struct.unpack(">d", struct.pack(">Q", ieee))[0]


def ieee_to_ibm(ieee):
    """
    Convert Python floating point numbers to IBM-format (bytes).
    """
    # Python uses IEEE: sign * 1.mantissa * 2 ** (exponent - 1023)
    # IBM mainframe:    sign * 0.mantissa * 16 ** (exponent - 64)

    if ieee == 0.0:
        return b'\x00' * 8

    # The IBM hexadecimal floating point (HFP) format represents the number
    # zero with all zero bits.  All zero bits is the "true zero" or normalized
    # form of zero.  Any values for the sign and exponent can be used if the
    # mantissa portion of the encoding is all zero bits, but an IBM machine
    # might lose precision when performing arithmetic with alternative zero
    # representations.  With that in mind, and because this format was not
    # defined with a mechanism for not-a-number (NaN) values, SAS uses
    # alternative zero encodings to represent NaN.  By default, a SAS missing
    # value is encoded with an ASCII-encoded period (".") as the first byte.

    if isinstance(ieee, xport.NaN):
        return bytes(ieee)
    if math.isnan(ieee):
        return b'.' + b'\x00' * 7
    if math.isinf(ieee):
        raise NotImplementedError('Cannot convert infinity')

    bits = struct.pack('>d', ieee)
    ulong, = struct.unpack('>Q', bits)

    sign = (ulong & (1 << 63)) >> 63  # 1-bit     sign
    exponent = ((ulong & (0x7ff << 52)) >> 52) - 1023  # 11-bits   exponent
    mantissa = ulong & 0x000fffffffffffff  # 52-bits   mantissa/significand

    if exponent > 248:
        msg = 'Cannot store magnitude more than ~ 16 ** 63 as IBM-format'
        raise Overflow(msg)
    if exponent < -260:
        msg = 'Cannot store magnitude less than ~ 16 ** -65 as IBM-format'
        raise Underflow(msg)

    # IEEE mantissa has an implicit 1 left of the radix:    1.significand
    # IBM mantissa has an implicit 0 left of the radix:     0.significand
    # We must bitwise-or the implicit 1.mmm into the mantissa
    # later we will increment the exponent to account for this change
    mantissa = 0x0010000000000000 | mantissa

    # IEEE exponents are for base 2:    mantissa * 2 ** exponent
    # IBM exponents are for base 16:    mantissa * 16 ** exponent
    # We must divide the exponent by 4, since 16 ** x == 2 ** (4 * x)
    quotient, remainder = divmod(exponent, 4)
    exponent = quotient

    # We don't want to lose information;
    # the remainder from the divided exponent adjusts the mantissa
    mantissa <<= remainder

    # Increment exponent, because of earlier adjustment to mantissa
    # this corresponds to the 1.mantissa vs 0.mantissa implicit bit
    exponent += 1

    # IBM exponents are excess 64
    exponent += 64

    # IBM has 1-bit sign, 7-bits exponent, and 56-bits mantissa.
    # We must shift the sign and exponent into their places.
    sign <<= 63
    exponent <<= 56

    # We lose some precision, but who said floats were perfect?
    return struct.pack('>Q', sign | exponent | mantissa)


def load(fp):
    """
    Deserialize a SAS dataset library from a SAS Transport v5 (XPT) file.

        >>> with open('test/data/example.xpt', 'rb') as f:
        ...     library = load(f)
    """
    try:
        bytestring = fp.read()
    except UnicodeDecodeError:
        raise TypeError(f'Expected a BufferedReader in bytes-mode, got {type(fp).__name__}')
    return loads(bytestring)


def loads(bytestring):
    """
    Deserialize a SAS dataset library from an XPORT-format string.

        >>> with open('test/data/example.xpt', 'rb') as f:
        ...     bytestring = f.read()
        >>> library = loads(bytestring)
    """
    return Library.from_bytes(bytestring)


def dump(library, fp):
    """
    Serialize a SAS dataset library to a SAS Transport v5 (XPORT) file.

        >>> library = Library()
        >>> with open('test/data/doctest.xpt', 'wb') as f:
        ...     dump(library, f)

    The input ``library`` can be either an ``xport.Library``, an
    ``xport.Dataset`` collection, or a single ``pandas.DataFrame``.
    An ``xport.Dataset`` is preferable, because that can be assigned a
    name, which SAS expects.

        >>> ds = xport.Dataset(name='EMPTY')
        >>> with open('test/data/doctest.xpt', 'wb') as f:
        ...     dump(ds, f)

    """
    fp.write(dumps(library))


def dumps(library):
    """
    Serialize a SAS dataset library to a string in XPORT-format.

        >>> library = Library()
        >>> bytestring = dumps(library)

    The input ``library`` can be either an ``xport.Library``, an
    ``xport.Dataset`` collection, or a single ``pandas.DataFrame``.
    An ``xport.Dataset`` is preferable, because that can be assigned a
    name, which SAS expects.

        >>> ds = xport.Dataset(name='EMPTY')
        >>> bytestring = dumps(ds)

    """
    return bytes(Library(library))



#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
"""
To upload to PyPI:

    $ python setup.py sdist
    $ python setup.py bdist_wheel --universal
    $ twine upload dist/*

"""
# Community Packages
from setuptools import setup

# Most arguments for ``setup`` should be written in ``setup.cfg``.
# https://setuptools.readthedocs.io/en/latest/setuptools.html#using-a-src-layout
setup()



"""
Read and write SAS XPORT/XPT-format files.
"""

# Standard Library
import enum
import logging
import re
import string
import struct
import textwrap
import warnings
from collections.abc import Mapping, MutableMapping
from datetime import datetime
from io import StringIO

# Community Packages
import pandas as pd

from .__about__ import __version__  # noqa: F401 module imported but unused

LOG = logging.getLogger(__name__)

__all__ = [
    'Library',
    'Member',
    'NaN',
]


class SpecialMissingValue(float):
    """
    Special missing values.

    SAS supports 27 special missing values, allowing the categorization of
    missing data by tagging or labeling missing values using the letters A to Z
    or an underscore.
    """

    pattern = re.compile(b'^(?P<tag>[A-Z_])' + b'\x00' * 7 + b'$')

    def __bytes__(self):
        """
        XPORT-format byte string.
        """
        return self.name.encode('ascii') + b'\x00' * 7

    @classmethod
    def unpack(cls, bytestring):
        """
        Create a ``SpecialMissingValue`` from an XPORT-format bytestring.

        For compatibility with standard Python tools, ``xport.v56.ibm_to_ieee``
        creates regular float NaNs rather than special missing values.
        """
        match = cls.pattern.match(bytestring)
        return NaN(match['tag'].decode('ascii'))


NaN = enum.Enum(
    value='NaN',
    names={c: float('nan')
           for c in '_' + string.ascii_uppercase},
    module=__name__,
    type=SpecialMissingValue,
)
NaN.__doc__ = SpecialMissingValue.__doc__


class VariableType(enum.IntEnum):
    """
    SAS variables can be either Numeric or Character type.
    """
    NUMERIC = 1
    CHARACTER = 2


class FormatAlignment(enum.IntEnum):
    """
    SAS formats are either left- or right-aligned.
    """
    LEFT = 0
    RIGHT = 1


class Informat:
    """
    SAS variable informat.
    """

    pattern = re.compile(
        r'^(?P<name>\$?[A-Z0-9]*?)(?P<w>\d+)\.(?P<d>\d+)?$',
        re.IGNORECASE,
    )

    #    char8 niform;      /* NAME OF INPUT FORMAT                   */
    #    short nifl;        /* INFORMAT LENGTH ATTRIBUTE              */
    #    short nifd;        /* INFORMAT NUMBER OF DECIMALS            */
    byte_structure = '>8shh'

    def __init__(self, name='', length=0, decimals=0):
        """
        Initialize an input format.
        """
        self._name = name
        self._length = length
        self._decimals = decimals

    def __str__(self):
        """
        Pleasant display value.
        """
        if not (self.name or self.length or self.decimals):
            return ''
        decimals = self.decimals if self.decimals else ''
        return f'{self.name}{self.length}.{decimals}'

    def __repr__(self):
        """
        REPL-format string.
        """
        return '{cls}(name={name!r}, length={length!r}, decimals={decimals!r})'.format(
            cls=type(self).__name__,
            name=self.name,
            length=self.length,
            decimals=self.decimals,
        )

    def __bytes__(self):
        """
        XPORT-format byte string.
        """
        fmt = self.byte_structure
        name = self.name.encode('ascii').ljust(8)
        if len(name) > 8:
            raise ValueError('ASCII-encoded {name!r} longer than 8 bytes')
        return struct.pack(fmt, name, self.length, self.decimals)

    @classmethod
    def unpack(cls, bytestring):
        """
        Create an informat from an XPORT-format bytestring.
        """
        fmt = cls.byte_structure
        return cls.from_struct_tokens(*struct.unpack(fmt, bytestring))

    @classmethod
    def from_struct_tokens(cls, name, length, decimals):
        """
        Create an informat from unpacked struct tokens.
        """
        name = name.strip(b'\x00').decode('ascii').strip()
        return cls(name=name, length=length, decimals=decimals)

    @classmethod
    def from_spec(cls, spec, *args, **kwds):
        """
        Create an informat from a text specification.
        """
        mo = cls.pattern.fullmatch(spec)
        if mo is None:
            raise ValueError(f'Invalid informat {spec}')
        name = mo['name'].upper()
        bytestring = name.encode('ascii')
        if len(bytestring) > 8:
            raise ValueError(f'ASCII-encoded name {bytestring} longer than 8 characters')
        length = int(mo.group('w'))
        try:
            decimals = int(mo.group('d'))
        except (TypeError, IndexError):
            decimals = 0
        form = cls(*args, name=name, length=length, decimals=decimals, **kwds)
        LOG.debug(f'Parsed {form!r} from {spec!r}')
        return form

    @property
    def name(self):
        """The name of the format."""  # noqa: D401
        return self._name

    @property
    def length(self):
        """The width value of the format: ``INFORMATw.``."""  # noqa: D401
        return self._length

    @property
    def decimals(self):
        """The ``d`` value of numeric formats: ``INFORMATw.d``."""  # noqa: D401
        # The documentation states that ``d`` optionally specifies the
        # power of 10 by which to divide numeric input.  If the data
        # contain decimal points, the ``d``` value is ignored.
        return self._decimals

    def __eq__(self, other):
        """Equality."""
        if not isinstance(other, Informat):
            raise TypeError(f"Can't compare {type(self).__name__} with {type(other).__name__}")
        attributes = [
            'name',
            'length',
            'decimals',
        ]
        return all(getattr(self, a) == getattr(other, a) for a in attributes)


class Format(Informat):
    """
    SAS variable format.
    """

    #    char8 nform;       /* NAME OF FORMAT                         */
    #    short nfl;          /* FORMAT FIELD LENGTH OR 0               */
    #    short nfd;         /* FORMAT NUMBER OF DECIMALS              */
    #    short nfj;         /* 0=LEFT JUSTIFICATION, 1=RIGHT JUST     */
    byte_structure = '>8shhh'

    def __init__(self, name='', length=0, decimals=0, justify=FormatAlignment.LEFT):
        """
        Initialize a SAS variable format.
        """
        self._justify = justify
        super().__init__(name, length, decimals)

    def __repr__(self):
        """
        REPL-format string.
        """
        fmt = '{cls}(name={name!r}, length={length!r}, decimals={decimals!r}, justify={justify})'
        return fmt.format(
            cls=type(self).__name__,
            name=self.name,
            length=self.length,
            decimals=self.decimals,
            justify=self.justify,
        )

    def __bytes__(self):
        """
        XPORT-format byte string.
        """
        # TODO: It'd be nice to avoid copy-pasting code from parent.
        fmt = self.byte_structure
        name = self.name.encode('ascii').ljust(8)
        if len(name) > 8:
            raise ValueError('ASCII-encoded {name!r} longer than 8 bytes')
        length = self.length if self.length is not None else 0
        decimals = self.decimals if self.decimals is not None else 0
        return struct.pack(fmt, name, length, decimals, self.justify)

    @classmethod
    def from_struct_tokens(cls, name, length, decimals, justify):
        """
        Create a format from unpacked struct tokens.
        """
        form = super().from_struct_tokens(name, length, decimals)
        form._justify = justify
        return form

    @classmethod
    def from_spec(cls, spec, justify=FormatAlignment.LEFT):
        """
        Create a format from a text specification.
        """
        return super().from_spec(spec=spec, justify=justify)

    @property
    def justify(self):
        """
        Left- or right-alignment.
        """
        return self._justify

    def __eq__(self, other):
        """Equality."""
        return super().__eq__(other) and self.justify == other.justify


# The Pandas documentation suggests avoiding inheritance, but their
# other options for extending ``Series`` objects fall flat, because so
# many Pandas methods return new instances.  To carry variable metadata
# over to the new instances, we need to overrride the constructors.
# https://pandas.pydata.org/pandas-docs/stable/development/extending.html


class Variable(pd.Series):
    """
    SAS variable.

    ``Variable`` extends Pandas' ``Series``, adding SAS metadata.
    """

    _metadata = [
        'label',
        'width',
        'vtype',
        '_format',
        '_informat',
    ]

    def copy_metadata(self, other):
        """
        Copy metadata from another Variable.
        """
        # LOG.debug(f'Copying metadata from {other}')  # BUG: Causes infinite recursion!
        if isinstance(other, Variable):
            for name in self._metadata:
                value = getattr(self, name, None)
                if value is None:
                    value = getattr(other, name, None)
                object.__setattr__(self, name, value)

    def __repr__(self):
        """REPL-format."""
        metadata = (name.strip('_') for name in self._metadata)
        metadata = {name: getattr(self, name, None) for name in metadata}
        metadata = (f'{name}: {value}' for name, value in metadata.items() if value is not None)
        return f'{type(self).__name__}\n{super().__repr__()}\n{", ".join(metadata)}'

    def __init__(
        self,
        data=None,
        index=None,
        dtype=None,
        name=None,
        copy=False,
        fastpath=False,
        label=None,
        vtype=None,
        width=None,
        format=None,
        informat=None,
        **kwds,
    ):
        """
        Initialize SAS variable metadata.
        """
        # TODO: Consider validating that the name isn't blank.
        metadata = {
            'label': label,
            'vtype': vtype,
            'width': width,
            'format': format,
            'informat': informat,
        }
        super().__init__(data, index, dtype, name, copy, fastpath, **kwds)
        for name, value in metadata.items():
            if value is not None:
                setattr(self, name, value)
        self.copy_metadata(data)
        for name, value in metadata.items():
            setattr(self, name, getattr(self, name, value))
        LOG.debug(f'Initialized {self}')

    def __finalize__(self, other, method=None, **kwds):
        """
        Extend Series finalize to handle more methods.
        """
        self = super().__finalize__(other, method, **kwds)
        if method == 'concat':
            first, *rest = other.objs
            source = first
        else:
            source = other
        self.copy_metadata(source)
        LOG.debug(f'Finalized {self}')
        return self

    @property
    def _constructor(self):
        """
        Construct an instance with the same dimensions as the original.
        """
        return Variable

    @property
    def _constructor_expanddim(self):
        """
        Construct an instance with an extra dimension.

        For example, transforming a series into a dataframe.
        """
        return pd.DataFrame

    @property
    def format(self):
        """
        SAS variable format.
        """
        return self._format

    @format.setter
    def format(self, value):
        if value is None:
            self._format = None
        elif isinstance(value, Format):
            self._format = value
        else:
            self._format = Format.from_spec(value)

        if self.format and self.format.name.startswith('$'):
            self.vtype = VariableType.CHARACTER
        elif self.format and (self.format.name or self.format.decimals):
            self.vtype = VariableType.NUMERIC

    @property
    def informat(self):
        """
        SAS variable informat.
        """
        return self._informat

    @informat.setter
    def informat(self, value):
        if value is None:
            self._informat = None
        elif isinstance(value, Informat):
            self._informat = value
        else:
            self._informat = Informat.from_spec(value)

        if self.informat and self.informat.name.startswith('$'):
            self.vtype = VariableType.CHARACTER
        elif self.informat and (self.informat.name or self.informat.decimals):
            self.vtype = VariableType.NUMERIC


class Dataset(pd.DataFrame):
    """
    SAS data set.

    ``Dataset`` extends Pandas' ``DataFrame``, adding SAS metadata.
    """

    _metadata = [
        'name',
        'dataset_label',
        'dataset_type',
        'created',
        'modified',
        'sas_os',
        'sas_version',
    ]

    def copy_metadata(self, other):
        """
        Copy metadata from a Dataset or mapping of Variables.
        """
        # LOG.debug(f'Copying metadata from {other}')  # BUG: Causes infinite recursion!
        if isinstance(other, Dataset):
            for name in self._metadata:
                object.__setattr__(self, name, getattr(other, name, None))
        if isinstance(other, (Dataset, Mapping)):
            for k, v in self.items():
                try:
                    v.copy_metadata(other[k])
                except KeyError:
                    continue

    def __repr__(self):
        """REPL-format."""
        metadata = (name.strip('_') for name in self._metadata)
        metadata = {name: getattr(self, name, None) for name in metadata}
        metadata = (f'{name}: {value}' for name, value in metadata.items() if value)
        template = '''\
            {cls} {name}
            {variables_metadata}

            {super}
            {metadata}
        '''
        return textwrap.dedent(template).format(
            cls=type(self).__name__,
            name=self.name,
            super=super().__repr__(),
            metadata=', '.join(metadata),
            variables_metadata=self.contents,
        )

    @property
    def label(self):
        return self.dataset_label

    @label.setter
    def label(self, value):
        self.dataset_label = value

    def __init__(
        self,
        data=None,
        index=None,
        columns=None,
        dtype=None,
        copy=False,
        name=None,
        label=None,
        dataset_label=None,
        dataset_type=None,
        created=None,
        modified=None,
        sas_os=None,
        sas_version=None,
        **kwds,
    ):
        """
        Initialize SAS dataset metadata.
        """
        if dataset_label is not None:
            label = dataset_label
            warnings.warn(
                'Use ``label`` instead of ``dataset_label``',
                DeprecationWarning,
                stacklevel=2,
            )
        # TODO: Consider validating dataset type: {'DATA', 'VIEW', 'CATALOG'}.
        #       I think only 'DATA' is supported by the XPORT format.
        # TODO: Consider validating that the name isn't blank.
        metadata = {
            'name': name,
            'dataset_label': label,
            'created': created,
            'modified': modified,
            'sas_os': sas_os,
            'sas_version': sas_version,
            'dataset_type': dataset_type,
        }
        super().__init__(data=data, index=index, columns=columns, dtype=dtype, copy=copy, **kwds)
        for name, value in metadata.items():
            if value is not None:
                setattr(self, name, value)
        self.copy_metadata(data)
        for name, value in metadata.items():
            setattr(self, name, getattr(self, name, value))
        # LOG.debug(f'Initialized {self}')  # BUG: Causes infinite recursion!

    def __finalize__(self, other, method=None, **kwds):
        """
        Propagate metadata to a copy.
        """
        # TODO: Is the call to super redundant?
        self = super().__finalize__(other, method, **kwds)
        if method == 'concat':
            first, *rest = other.objs
            source = first
        elif method == 'merge':
            source = other.left
        else:
            source = other
        self.copy_metadata(source)
        LOG.debug(f'Finalized {self}')
        return self

    def __setitem__(self, key, value):
        """
        When inserting/updating a column, we must copy metadata.
        """
        # TODO: There are probably other ways Pandas adds columns to
        #       a DataFrame.  We need to copy metadata in those, too.
        old = self.iloc[:0].copy()
        super().__setitem__(key, value)
        if isinstance(value, Variable):
            self[key].copy_metadata(value)
        for k, v in old.items():
            if k != key:
                self[k].copy_metadata(v)

    @property
    def _constructor(self):
        """
        Construct an instance with the same dimensions as the original.
        """
        return Dataset

    @property
    def _constructor_sliced(self):
        """
        Construct an instance with one less dimension.

        For example, slicing a single column from a dataframe.
        """
        return Variable

    @property
    def contents(self):
        """
        Variable metadata, such as label, format, number, and position.
        """
        df = pd.DataFrame({
            'Variable': v.name,
            'Type': v.vtype.name.title() if v.vtype is not None else '',
            'Length': v.width,
            'Format': str(v.format) if v.format is not None else '',
            'Informat': str(v.informat) if v.informat is not None else '',
            'Label': v.label if v.label is not None else '',
        } for k, v in self.items())
        if df.empty:
            return df
        df.index = df.index + 1
        df.index.name = '#'
        # BUG: Pandas Series.cumsum() seems to fail on its new Int type;
        #      thus we have a redundant conversion to Int64Dtype.
        df['Position'] = df['Length'].cumsum().astype(pd.Int64Dtype())
        df['Length'] = df['Length'].fillna(pd.NA).astype(pd.Int64Dtype())
        df.loc[1, 'Position'] = 0
        return df

    def infos(self):
        """
        Like ``DataFrame.info`` but returns a string.
        """
        buf = StringIO()
        self.info(buf=buf)
        buf.seek(0)
        return buf.read()


class Library(MutableMapping):
    """
    Collection of datasets from a SAS file.
    """

    def __init__(self, members=(), created=None, modified=None, sas_os='', sas_version=''):
        """
        Initialize a SAS data library.
        """
        if created is None:
            created = datetime.now()
        if modified is None:
            modified = created
        self.created = created
        self.modified = modified
        self.sas_os = sas_os
        self.sas_version = sas_version

        # Convert a single dataset or dataframe to a collection of them.
        if isinstance(members, pd.DataFrame):
            members = {getattr(members, 'name', None): members}

        self._members = {}
        if isinstance(members, Library):
            self._members = members._members
            self.created = members.created
            self.modified = members.modified
            self.sas_os = members.sas_os
            self.sas_version = members.sas_version
        elif isinstance(members, Mapping):
            for name, dataset in members.items():
                self[name] = dataset  # Use __setitem__ to validate metadata.
        else:
            for dataset in members:
                if dataset.name in self:
                    warnings.warn(f'More than one dataset named {dataset.name!r}')
                self[dataset.name] = dataset

    def __repr__(self):
        """
        REPL-format string.
        """
        fmt = '<{cls} members={members}>'
        return fmt.format(cls=type(self).__name__, members=list(self))

    def __getitem__(self, name):
        """
        Get a member dataset.
        """
        return self._members[name]

    def __setitem__(self, name, dataset):
        """
        Insert or update a member in the library.
        """
        if not isinstance(dataset, Dataset):
            dataset = Dataset(dataset, name=name)
        elif dataset.name is None and name is not None:
            dataset.name = name
            warnings.warn(f'Set dataset name to {name!r}')
        elif name != dataset.name:
            raise ValueError(f'Library member name {name} must match dataset name {dataset.name}')
        self._members[name] = dataset

    def __delitem__(self, name):
        """
        Remove a member datset.
        """
        del self._members[name]

    def __iter__(self):
        """
        Get an iterator of dataset names.
        """
        return iter(self._members)

    def __len__(self):
        """
        Get the number of datasets in the library.
        """
        return len(self._members)

    def __eq__(self, other):
        """
        Compare equality.
        """
        same_keys = set(self) == set(other)
        same_values = all((self[k] == other[k]).all(axis=None) for k in self)
        return same_keys and same_values


########################################################################
# Legacy, keeping backwards compatibility.


def from_columns(mapping, fp):
    """
    Write columns to the open file opbject ``fp`` in XPT-format.

    The mapping should be of column names to equal-length sequences.

    Column labels are restricted to 40 characters. The XPT format also
    requires a separate column "name" that is restricted to 8
    characters. This name will be automatically created based on the
    column label -- the first 8 characters, non-alphabet characters
    replaced with underscores, padded to 8 characters if necessary.  All
    text strings, including column labels, will be converted to bytes
    using the ISO-8859-1 encoding.
    """
    df = pd.DataFrame(mapping)
    return from_dataframe(df, fp)


def from_rows(iterable, fp):
    """
    Write rows to the open file object ``fp`` in XPT-format.

    In this case, ``rows`` should be an iterable of iterables, such as a
    list of tuples. If the rows are mappings or namedtuples (or any
    instance of a tuple that has a ``._fields`` attribute), the column
    labels will be inferred from the keys or attributes of the first
    row.

    Column labels are restricted to 40 characters. The XPT format also
    requires a separate column "name" that is restricted to 8
    characters. This name will be automatically created based on the
    column label -- the first 8 characters, non-alphabet characters
    replaced with underscores, padded to 8 characters if necessary.  All
    text strings, including column labels, will be converted to bytes
    using the ISO-8859-1 encoding.
    """
    df = pd.DataFrame(iterable)
    df.columns = [f'x{i:02d}' if isinstance(i, int) else i for i in df]
    return from_dataframe(df, fp)


def from_dataframe(dataframe, fp):
    """
    Write a Pandas ``DataFrame`` to an XPORT-format file.

    Serialize to ``fp``, an open file-like object.
    """
    # Avoid circular import problems.
    # Xport Modules
    from xport.v56 import dump
    warnings.warn('Please use ``xport.v56.dump`` in the future', DeprecationWarning)
    library = Library([Dataset(dataframe)])
    dump(library, fp)


def to_rows(fp):
    """
    Read a file in XPT-format and return rows.

    Deserialize ``fp`` (a ``.read()``-supporting file-like object
    containing an XPT document) to a list of rows. As XPT files are
    encoded in their own special format, the ``fp`` object must be in
    bytes-mode. ``Row`` objects will be namedtuples with attributes
    parsed from the XPT metadata.
    """
    df = to_dataframe(fp)
    return list(df.itertuples(index=False, name=None))


def to_columns(fp):
    """
    Read a file in XPT-format and return columns as a dict of lists.

    Deserialize ``fp`` (a ``.read()``-supporting file-like object
    containing an XPT document) to a list of rows. As XPT files are
    encoded in their own special format, the ``fp`` object must be in
    bytes-mode.
    """
    dataset = to_dataframe(fp)
    return {k: v for k, v in dataset.items()}


def to_numpy(fp):
    """
    Read a file in SAS XPT format and return a NumPy array.

    Deserialize ``fp`` (a ``.read()``-supporting file-like object
    containing an XPT document) to a list of rows. As XPT files are
    encoded in their own special format, the ``fp`` object must be in
    bytes-mode.
    """
    return to_dataframe(fp).values


def to_dataframe(fp):
    """
    Read a file in SAS XPT format and return a Pandas DataFrame.

    Deserialize ``fp`` (a ``.read()``-supporting file-like object
    containing an XPT document) to a list of rows. As XPT files are
    encoded in their own special format, the ``fp`` object must be in
    bytes-mode.
    """
    # Avoid circular import problems.
    # Xport Modules
    from xport.v56 import load
    warnings.warn('Please use ``xport.v56.load`` in the future', DeprecationWarning)
    library = load(fp)
    dataset = next(iter(library.values()))
    return dataset


class Reader:
    """
    Read records from a SAS Transport (XPORT) file.

    Deserialize ``self._fp`` (a ``.read()``-supporting file-like object
    containing an XPT document) to a Python object.

    The returned object is an iterator.  Each iteration returns an
    observation from the XPT file.

        with open('example.xpt', 'rb') as f:
            for row in xport.Reader(f):
                process(row)
    """

    def __init__(self, fp):
        self.dataset = to_dataframe(fp)

    def __iter__(self):
        return iter(self.dataset.itertuples(index=False, name='Observation'))

    @property
    def fields(self):
        return tuple(self.dataset.columns)

    def __getattr__(self, name):
        return getattr(self.dataset, name)


class XportReader(Reader):
    pass  # Alias.  https://github.com/selik/xport/issues/55


class NamedTupleReader(Reader):
    pass  # Reader already yields namedtuples.


class DictReader(Reader):

    def __iter__(self):
        for row in super().__iter__():
            yield row._asdict()
